#!/usr/bin/python3
"""Tests for the templateparser module."""

# Too many public methods
# pylint: disable=R0904

# Standard modules
import os
import re
import time
import unittest

# Unittest target
from uweb3 import templateparser

class Parser(unittest.TestCase):
  """Basic tests for the Parser class and equality of Template objects."""
  def setUp(self):
    """Creates a template file and a similar instance attribute."""
    self.name = 'tmp_template'
    self.raw = 'This is a basic [noun]'
    self.template = templateparser.Template(self.raw)
    with open(self.name, 'w') as template:
      template.write('This is a basic [noun]')
      template.flush()

  def tearDown(self):
    """Removes the template file from the filesystem."""
    os.unlink('tmp_template')

  def testAddTemplate(self):
    """[Parser] AddTemplate adds a template to the parser"""
    parser = templateparser.Parser()
    self.assertEqual(len(parser), 0)
    parser.AddTemplate(self.name)
    self.assertEqual(len(parser), 1)
    self.assertEqual(parser[self.name], self.template)

  def testAccessTemplate(self):
    """[Parser] getting a template by key loads it when required"""
    parser = templateparser.Parser()
    self.assertEqual(len(parser), 0)
    self.assertEqual(parser[self.name], self.template)
    self.assertEqual(len(parser), 1)

  def testOverWriteTemplate(self):
    """[Parser] AddTemplate overrides previously loaded template"""
    custom_raw = 'My very own [adj] template'
    custom_tmpl = templateparser.Template(custom_raw)
    parser = templateparser.Parser()
    parser.AddTemplate(self.name)
    # Create a new template in place of the existing one, and reload it.
    with open(self.name, 'w') as tmpl:
      tmpl.write(custom_raw)
      tmpl.flush()
    # Assert the template has not yet changed, load it, assert that is has.
    self.assertNotEqual(custom_tmpl, parser[self.name])
    parser.AddTemplate(self.name)
    self.assertEqual(parser[self.name], custom_tmpl)

  def testPreloadTemplates(self):
    """[Parser] Templates can be preloaded when instantiating the Parser"""
    parser = templateparser.Parser(templates=[self.name])
    self.assertEqual(len(parser), 1)
    self.assertEqual(parser[self.name], self.template)

  def testParseVersusParseString(self):
    """[Parser] Parse and ParseString only differ in cached lookup"""
    parser = templateparser.Parser()
    result_parse = parser[self.name].Parse()
    result_parse_string = parser.ParseString(self.raw)
    self.assertEqual(result_parse, result_parse_string)


class ParserPerformance(unittest.TestCase):
  """Basic performance test of the Template's initialization and Parsing."""
  @staticmethod
  def testPerformance():
    """[Parser] Basic performance test for 2 template replacements"""
    template = 'This [obj:foo] is just a quick [bar]'
    for _template in range(100):
      tmpl = templateparser.Template(template)
      for _parse in range(100):
        tmpl.Parse(obj={'foo': 'template'}, bar='hack')


class TemplateTagBasic(unittest.TestCase):
  """Tests validity and parsing of simple tags."""
  def setUp(self):
    """Makes the Template class available on the instance."""
    self.tmpl = templateparser.Template

  def testTaglessTemplate(self):
    """[BasicTag] Templates without tags get returned verbatim as SafeString"""
    template = 'Template without any tags'
    self.assertEqual(self.tmpl(template).Parse(), template)

  def testSafeString(self):
    """[BasicTag] Templates without tags get returned verbatim as SafeString"""
    template = 'Template without any tags'
    parsed_template = self.tmpl(template).Parse()
    self.assertTrue(isinstance(parsed_template, templateparser.HTMLsafestring))

  def testSingleTagTemplate(self):
    """[BasicTag] Templates with basic tags get returned proper"""
    template = 'Template with [single] tag'
    result = self.tmpl(template).Parse(single='just one')
    self.assertEqual(result, 'Template with just one tag')

  def testSaveTagTemplate(self):
    """[BasicTag] Templates with basic tags get returned properly when replacement is already html safe"""
    template = 'Template with just [single] tag'
    result = self.tmpl(template).Parse(single=templateparser.HTMLsafestring('<b>a safe</b>'))
    self.assertEqual(result, 'Template with just <b>a safe</b> tag')

  def testUnsaveTagTemplate(self):
    """[BasicTag] Templates with basic tags get returned properly when replacement is not html safe"""
    template = 'Template with just [single] tag'
    result = self.tmpl(template).Parse(single='<b>an unsafe</b>')
    self.assertEqual(result, 'Template with just &lt;b&gt;an unsafe&lt;/b&gt; tag')

  def testCasedTag(self):
    """[BasicTag] Tag names are case-sensitive"""
    template = 'The parser has no trouble with [cAsE] [case].'
    result = self.tmpl(template).Parse(cAsE='mixed')
    self.assertEqual(result, 'The parser has no trouble with mixed [case].')

  def testUnderscoredTag(self):
    """[BasicTag] Tag names may contain underscores"""
    template = 'The template may contain [under_scored] tags.'
    result = self.tmpl(template).Parse(under_scored='underscored')
    self.assertEqual(result, 'The template may contain underscored tags.')

  def testMultiTagTemplate(self):
    """[BasicTag] Multiple instances of a tag will all be replaced"""
    template = '[adjective] [noun] are better than other [noun].'
    result = self.tmpl(template).Parse(noun='cows', adjective='Beefy')
    self.assertEqual(result, 'Beefy cows are better than other cows.')

  def testEmptyOrWhitespace(self):
    """[BasicTag] Empty tags or tags containing whitespace aren't actual tags"""
    template = 'This [is a] broken [] template, really'
    result = self.tmpl(template).Parse(**{'is a': 'HORRIBLY', '': ', NASTY'})
    self.assertEqual(result, template)

  def testBadCharacterTags(self):
    """[BasicTag] Tags containing bad characters are not considered tags"""
    bad_chars = """ :~!@#$%^&*()+-={}\|;':",./<>? """
    template = ''.join('[%s] [check]' % char for char in bad_chars)
    expected = ''.join('[%s] ..' % char for char in bad_chars)
    replaces = {char: 'FAIL' for char in bad_chars}
    replaces['check'] = '..'
    self.assertEqual(self.tmpl(template).Parse(**replaces), expected)

  def testUnreplacedTag(self):
    """[BasicTag] Template tags without replacement are returned verbatim"""
    template = 'Template with an [undefined] tag.'
    self.assertEqual(self.tmpl(template).Parse(), template)

  def testUnreplacedTag(self):
    """[BasicTag] Access to private members is not allowed"""
    template = 'Template with an [private.__class__] tag.'
    self.assertEqual(self.tmpl(template).Parse(), template)

  def testBracketsInsideTag(self):
    """[BasicTag] Innermost bracket pair are the tag's delimiters"""
    template = 'Template tags may not contain [[spam][eggs]].'
    expected = 'Template tags may not contain [opening or closing brackets].'
    result = self.tmpl(template).Parse(
        **{'[spam': 'EPIC', 'eggs]': 'FAIL', 'spam][eggs': 'EPIC FAIL',
           'spam': 'opening or ', 'eggs': 'closing brackets'})
    self.assertEqual(result, expected)

  def testTemplateInterpolationSyntax(self):
    """[BasicTag] Templates support string interpolation of dicts"""
    template = 'Hello [name]'
    self.assertEqual(self.tmpl(template) % {'name': 'Bob'}, 'Hello Bob')


class TemplateTagIndexed(unittest.TestCase):
  """Tests the handling of complex tags (those with attributes/keys/indexes)."""
  def setUp(self):
    """Sets up a parser instance, as it never changes."""
    self.tmpl = templateparser.Template

  def testTemplateMappingKey(self):
    """[IndexedTag] Template tags can address mappings properly"""
    template = 'This uses a [dictionary:key].'
    result = self.tmpl(template).Parse(dictionary={'key': 'spoon'})
    self.assertEqual(result, 'This uses a spoon.')

  def testTemplateIndexing(self):
    """[IndexedTag] Template tags can access indexed iterables"""
    template = 'Template that grabs the [obj:2] key from the given tuple/list.'
    expected = 'Template that grabs the third key from the given tuple/list.'
    numbers = 'first', 'second', 'third'
    self.assertEqual(self.tmpl(template).Parse(obj=numbers), expected)
    self.assertEqual(self.tmpl(template).Parse(obj=list(numbers)), expected)

  def testTemplateAttributes(self):
    """[IndexedTag] Template tags will do attribute lookups after key-lookups"""
    class Mapping(dict):
      """A subclass of a dictionary, so we can define attributes on it."""
      NAME = 'attribute'

    template = 'Template used [tag:NAME] lookup.'
    lookup_attr = 'Template used attribute lookup.'
    lookup_dict = 'Template used key (mapping) lookup.'

    mapp = Mapping()
    self.assertEqual(self.tmpl(template).Parse(tag=mapp), lookup_attr)
    mapp['NAME'] = 'key (mapping)'
    self.assertEqual(self.tmpl(template).Parse(tag=mapp), lookup_dict)

  def testTemplateIndexingCharacters(self):
    """[IndexedTag] Tags indexes may be made of word chars and dashes only,
    they should however not start and end with _ to avoid access to
    private vars.
    _ is allowed elsewhere in the string."""
    good_chars = "aAzZ0123-"
    bad_chars = """ :~!@#$%^&*()+={}\|;':",./<>? """
    for index in good_chars:
      tag = {index: 'SUCCESS'}
      template = '[tag:%s]' % index
      self.assertEqual(self.tmpl(template).Parse(tag=tag), 'SUCCESS')
    for index in bad_chars:
      tag = {index: 'FAIL'}
      template = '[tag:%s]' % index
      self.assertEqual(self.tmpl(template).Parse(tag=tag), template)

  def testTemplateUnderscoreCharacters(self):
    """[IndexedTag] Tags indexes may be made of word chars and dashes only,
    they should however not start and end with _ to avoid access to
    private vars.
    _ is allowed elsewhere in the string."""
    # see if objects with underscores are reachable
    tag = {'test_test': 'SUCCESS'}
    template = '[tag:%s]' % 'test_test'
    self.assertEqual(self.tmpl(template).Parse(tag=tag), 'SUCCESS')

    tag = {'_test': 'SUCCESS'}
    template = '[tag:%s]' % '_test'
    self.assertEqual(self.tmpl(template).Parse(tag=tag), 'SUCCESS')

    tag = {'test_': 'SUCCESS'}
    template = '[tag:%s]' % 'test_'
    self.assertEqual(self.tmpl(template).Parse(tag=tag), 'SUCCESS')

    # check if private vars are impossible to reach.
    tag = {'_test_': 'SUCCESS'}
    template = '[tag:%s|raw]' % '_test_'
    self.assertEqual(self.tmpl(template).Parse(tag=tag), repr(tag))

  def testTemplateMissingIndexes(self):
    """[IndexedTag] Tags with bad indexes will be returned verbatim"""
    class Object:
      """A simple object to store an attribute on."""
      NAME = 'Freeman'

    template = 'Hello [titles:1] [names:NAME], how is [names:other] [date:now]?'
    expected = 'Hello [titles:1] Freeman, how is [names:other] [date:now]?'
    result = self.tmpl(template).Parse(titles=['Mr'], names=Object(), date={})
    self.assertEqual(result, expected)

  def testTemplateMultipleIndexing(self):
    """[IndexedTag] Template tags can contain multiple nested indexes"""
    template = 'Welcome to the [foo:bar:zoink].'
    result = self.tmpl(template).Parse(foo={'bar': {'zoink': 'World'}})
    self.assertEqual(result, 'Welcome to the World.')


class TemplateTagFunctions(unittest.TestCase):
  """Tests the functions that are performed on replaced tags."""
  def setUp(self):
    """Sets up a parser instance, as it never changes."""
    self.parser = templateparser.Parser()
    self.parse = self.parser.ParseString

  def testBasicFunction(self):
    """[TagFunctions] and html safe output"""
    template = 'This function does [none].'
    result = self.parse(template, none='"nothing"')
    self.assertEqual(result, 'This function does &quot;nothing&quot;.')

  def testBasicFunctionNumeric(self):
    """[TagFunctions] and html safe output for non string outputs"""
    template = '[tag]'
    result = self.parse(template, tag=1)
    self.assertEqual(result, '1')

  def testBasicFunctionRaw(self):
    """[TagFunctions] Raw function does not affect output"""
    template = 'This function does [none|raw].'
    result = self.parse(template, none='"nothing"')
    self.assertEqual(result, 'This function does "nothing".')

  def testNonexistantFuntion(self):
    """[TagFunctions] An error is raised for functions that don't exist"""
    template = 'This tag function is missing [num|zoink].'
    self.assertEqual(self.parse(template), template)
    # Error is only thrown if we actually pass an argument for the tag:
    self.assertRaises(templateparser.TemplateFunctionError,
                      self.parse, template, num=1)

  def testAlwaysString(self):
    """[TagFunctions] Tag function return is always converted to string."""
    template = '[number]'
    self.assertEqual(self.parse(template, number=1), '1')
    template = '[number|raw]'
    self.assertEqual(self.parse(template, number=2), '2')
    template = '[number|int]'
    self.parser.RegisterFunction('int', int)
    self.assertEqual(self.parse(template, number=3), '3')

  def testFunctionCharacters(self):
    """[TagFunctions] Tags functions may contain word chars and dashes only"""
    good_funcs = "aAzZ0123-_"
    good_func = lambda tag: 'SUCCESS'
    bad_funcs = """ :~!@#$%^&*+={}\;':"./<>?| """
    bad_func = lambda tag: 'FAIL'
    for index in good_funcs:
      template = '[tag|%s]' % index
      self.parser.RegisterFunction(index, good_func)
      self.assertEqual(self.parse(template, tag='foo'), 'SUCCESS')
    for index in bad_funcs:
      template = '[tag|%s]' % index
      self.parser.RegisterFunction(index, bad_func)
      self.assertEqual(self.parse(template, tag='foo'), template)
    self.parser.RegisterFunction('|', bad_func)

  def testDefaultHtmlSafe(self):
    """[TagFunctions] The default function escapes HTML entities"""
    default = 'This function does [none].'
    escaped = 'This function does [none|html].'
    expected = 'This function does &quot;nothing&quot;.'
    self.assertEqual(self.parse(default, none='"nothing"'), expected)
    self.assertEqual(self.parse(escaped, none='"nothing"'), expected)

  def testNoDefaultForSafeString(self):
    """[TagFunctions] The default function does not act upon SafeString parts"""
    first_template = 'Hello doctor [name]'
    second_template = '<assistant> [quote].'
    result = '<assistant> Hello doctor &quot;Who&quot;.'
    result_first = self.parse(first_template, name='"Who"')
    result_second = self.parse(second_template, quote=result_first)
    self.assertEqual(result, result_second)

  def testCustomFunction(self):
    """[TagFunctions] Custom functions can be added to the Parser"""
    self.parser.RegisterFunction('twice', lambda x: x + ' ' + x)
    template = 'The following will be stated [again|twice].'
    result = 'The following will be stated twice twice.'
    self.assertEqual(result, self.parse(template, again='twice'))

  def testFunctionChaining(self):
    """[TagFunctions] Multiple functions can be chained after one another"""
    self.parser.RegisterFunction('count', lambda x: '%s characters' % x)
    template = 'A replacement processed by two functions: [spam|len|count].'
    result = 'A replacement processed by two functions: 8 characters.'
    self.assertEqual(result, self.parse(template, spam='ham&eggs'))

  def testFunctionUse(self):
    """[TagFunctions] Tag functions are only called when requested by tags"""
    fragments_received = []
    def CountAndReturn(fragment):
      """Returns the given fragment after adding it to a counter list."""
      fragments_received.append(fragment)
      return fragment

    self.parser.RegisterFunction('count', CountAndReturn)
    template = 'Count only has [num|count] call, or it is [noun|raw].'
    result = self.parse(template, num='one', noun='broken')
    self.assertEqual(result, 'Count only has one call, or it is broken.')
    self.assertEqual(len(fragments_received), 1)

  def testTagFunctionUrl(self):
    """[TagFunctions] The tag function 'url' is present and works"""
    template = 'http://example.com/?breakfast=[query|url]'
    result = self.parse(template, query='"ham & eggs"')
    self.assertEqual(result, 'http://example.com/?breakfast=%22ham+%26+eggs%22')

  def testTagFunctionItems(self):
    """[TagFunctions] The tag function 'items' is present and works"""
    template = '[tag|items|raw]'
    tag = {'ham': 'eggs'}
    result = "[('ham', 'eggs')]"
    self.assertEqual(result, self.parse(template, tag=tag))

  def testTagFunctionValues(self):
    """[TagFunctions] The tag function 'values' is present and works"""
    template = '[tag|values|raw]'
    self.assertEqual(self.parse(template, tag={'ham': 'eggs'}), "['eggs']")

  def testTagFunctionSorted(self):
    """[TagFunctions] The tag function 'sorted' is present and works"""
    template = '[numbers|sorted]'
    numbers = [5, 1, 3, 2, 4]
    self.assertEqual(self.parse(template, numbers=numbers), "[1, 2, 3, 4, 5]")

  def testTagFunctionLen(self):
    """[TagFunctions] The tag function 'len' is present and works"""
    template = '[numbers|len]'
    self.assertEqual(self.parse(template, numbers=range(12)), "12")


class TemplateTagFunctionClosures(unittest.TestCase):
  """Tests the functions that are performed on replaced tags."""
  @staticmethod
  def Limit(length=80):
    """Returns a closure that limits input to a number of chars/elements."""
    return lambda string: string[:length]

  @staticmethod
  def LimitString(length=80, endchar='...'):
    """Limits input to `length` chars and appends `endchar` if it was longer."""
    def _Limit(string, length=length, endchar=endchar):
      if len(string) > length:
        return string[:length] + endchar
      return string
    return _Limit

  def setUp(self):
    """Sets up a parser instance, as it never changes."""
    self.parser = templateparser.Parser()
    self.parser.RegisterFunction('limit', self.Limit)
    self.parser.RegisterFunction('strlimit', self.LimitString)
    self.parse = self.parser.ParseString
    self.tag = 'hello world ' * 10

  def testSimpleClosureWithoutArguments(self):
    """[TagClosures] Simple tag closure-functions without arguments succeed"""
    template = '[tag|limit()]'
    result = self.parse(template, tag=self.tag)
    self.assertEqual(result, self.tag[:80])

  def testSimpleClosureArgument(self):
    """[TagClosures] Simple tag-closure functions operate on their argument"""
    template = '[tag|limit(20)]'
    result = self.parse(template, tag=self.tag)
    self.assertEqual(result, self.tag[:20])

  def testMathClosureArgument(self):
    """[TagClosures] Math tag-closure functions operate on their argument"""
    template = '[tag|limit(5*4)]'
    result = self.parse(template, tag=self.tag)
    self.assertEqual(result, self.tag[:20])

  def testFunctionClosureArgument(self):
    """[TagClosures] tags that use function calls in their function input should
    never be parsed"""
    template = '[tag|limit(abs(-20))]'
    result = self.parse(template, tag=self.tag)
    self.assertEqual(result, template)

  def testVariableClosureArgument(self):
    """[TagClosures] tags that try to use vars in their function arguments
    should never have access to the python scope."""
    test = 20
    template = '[tag|limit(test)]'
    self.assertRaises(templateparser.TemplateNameError,
        self.parse, template, tag=self.tag)

  def testComplexClosureWithoutArguments(self):
    """[TagClosures] Complex tag closure-functions without arguments succeed"""
    template = '[tag|strlimit()]'
    result = self.parse(template, tag=self.tag)
    self.assertEqual(len(result), 83)
    self.assertEqual(result[:80], self.tag[:80])
    self.assertEqual(result[-3:], '...')

  def testComplexClosureArguments(self):
    """[TagClosures] Complex tag closure-functions operate on arguments"""
    template = '[tag|strlimit(20, "TOOLONG")]'
    result = self.parse(template, tag=self.tag)
    self.assertEqual(len(result), 27)
    self.assertEqual(result[:20], self.tag[:20])
    self.assertEqual(result[-7:], 'TOOLONG')

  def testCharactersInClosureArguments(self):
    """[TagClosures] Arguments strings may contain specialchars"""
    template = '[tag|strlimit(20, "`-=./<>?`!@#$%^&*_+[]\{}|;\':")|raw]'
    result = self.parser.ParseString(template, tag=self.tag)
    self.assertTrue(result.endswith('`-=./<>?`!@#$%^&*_+[]\{}|;\':'))

  def testCommaInArgument(self):
    """[TagClosures] String arguments may contain commas"""
    template = '[tag|strlimit(10, "ham, eggs")]'
    result = self.parse(template, tag=self.tag)
    self.assertEqual(result[-9:], "ham, eggs")

  def testNamedArguments(self):
    """[TagClosures] Named arguments are not allowed"""
    template = '[tag|limit(length=20)]'
    self.assertRaises(templateparser.TemplateSyntaxError,
                      self.parse, template, tag=self.tag)

  def testTrailingComma(self):
    """[TagClosures] Arguments may not have a trailing comma"""
    template = '[tag|limit(20,)]'
    self.assertRaises(templateparser.TemplateSyntaxError,
                      self.parse, template, tag=self.tag)


class TemplateUnicodeSupport(unittest.TestCase):
  """TemplateParser handles Unicode gracefully."""
  def setUp(self):
    """Sets up a parser instance, as it never changes."""
    self.parser = templateparser.Parser()
    self.parse = self.parser.ParseString

  def testTemplateUnicode(self):
    """[Unicode] Templates may contain raw Unicode codepoints"""
    # And they will be converted to UTF8 eventually
    template = u'We \u2665 Python'
    self.assertEqual(self.parse(template), template)

  def testTemplateUTF8(self):
    """[Unicode] Templates may contain UTF8 encoded text"""
    # That is, input bytes will be left untouched
    template = u'We \u2665 Python'
    self.assertEqual(self.parse(template), template)

  def testUnicodeReplacements(self):
    """[Unicode] Unicode in tag replacements is converted to UTF8"""
    template = 'Underdark Web framework, also known as [name].'
    expected = u'Underdark Web framework, also known as \xb5Web.'
    self.assertEqual(self.parse(template, name=u'\xb5Web'), expected)

  def testUnicodeTagFunction(self):
    """[Unicode] Template functions returning unicode are converted to UTF8"""
    function_result = u'No more \N{BLACK HEART SUIT}'
    def StaticReturn(_fragment):
      """Returns a static string, for any input fragment."""
      return function_result

    self.parser.RegisterFunction('nolove', StaticReturn)
    template = '[love|nolove]'
    expected = function_result
    self.assertEqual(self.parse(template, love='love'), expected)

  def testTemplateTagUTF8(self):
    """[Unicode] Template tags may contain UTF8"""
    template = u'We \u2665 \xb5Web!'
    self.assertEqual(self.parse(template), template)


class TemplateInlining(unittest.TestCase):
  """TemplateParser properly handles the include statement."""
  def setUp(self):
    """Sets up a testbed."""
    self.parser = templateparser.Parser()
    self.parse = self.parser.ParseString
    self.tmpl = templateparser.Template

  def testInlineExisting(self):
    """{{ inline }} Parser will inline an already existing template reference"""
    self.parser['template'] = self.tmpl('This is a subtemplate by [name].')
    template = '{{ inline template }}'
    expected = 'This is a subtemplate by Elmer.'
    self.assertEqual(self.parse(template, name='Elmer'), expected)

  def testInlineFile(self):
    """{{ inline }} Parser will load an inlined template from file if needed"""
    with open('tmp_template', 'w') as inline_file:
      inline_file.write('This is a subtemplate by [name].')
      inline_file.flush()
    try:
      template = '{{ inline tmp_template }}'
      expected = 'This is a subtemplate by Elmer.'
      self.assertEqual(self.parse(template, name='Elmer'), expected)
    finally:
      os.unlink('tmp_template')


class TemplateConditionals(unittest.TestCase):
  """TemplateParser properly handles if/elif/else statements."""
  def setUp(self):
    """Sets up a testbed."""
    self.parse = templateparser.Parser().ParseString

  def testBasicConditional(self):
    """{{ if }} Basic boolean check works for relevant data types"""
    template = '{{ if [variable] }} ack {{ endif }}'
    # Boolean True inputs should return a SafeString object stating 'foo'.
    self.assertEqual(self.parse(template, variable=True), ' ack ')
    self.assertEqual(self.parse(template, variable='truth'), ' ack ')
    self.assertEqual(self.parse(template, variable=12), ' ack ')
    self.assertEqual(self.parse(template, variable=[1, 2]), ' ack ')
    # Boolean False inputs should yield an empty SafeString object.
    self.assertFalse(self.parse(template, variable=None))
    self.assertFalse(self.parse(template, variable=0))
    self.assertFalse(self.parse(template, variable=''))

  def testCompareTag(self):
    """{{ if }} Basic tag value comparison"""
    template = '{{ if [variable] == 5 }} foo {{ endif }}'
    self.assertFalse(self.parse(template, variable=0))
    self.assertFalse(self.parse(template, variable=12))
    self.assertTrue(self.parse(template, variable=5))

  def testCompareMath(self):
    """{{ if }} Basic math"""
    template = '{{ if 5*5 == 25 }} foo {{ endif }}'
    self.assertEqual(self.parse(template, variable=5).strip(), 'foo')

  def testTagIsInstance(self):
    """{{ if }} Tag value after python function comparison"""
    template = '{{ if isinstance([variable], int) }} ack {{ endif }}'
    self.assertFalse(self.parse(template, variable=[1]))
    self.assertFalse(self.parse(template, variable='number'))
    self.assertEqual(self.parse(template, variable=5), ' ack ')

  def testComparePythonFunction(self):
    """{{ if }} Tag value after python len comparison"""
    template = '{{ if len([variable]) == 5 }} foo {{ endif }}'
    self.assertEqual(self.parse(template, variable=[1,2,3,4,5]).strip(), 'foo')

  def testCompareNotallowdPythonFunction(self):
    """{{ if }} Tag value after python len comparison"""
    template = '{{ if open([variable]) == 5 }} foo {{ endif }}'
    self.assertRaises(templateparser.TemplateEvaluationError, self.parse, template)

  def testDefaultElse(self):
    """{{ if }} Else block will be parsed when `if` fails"""
    template = '{{ if [var] }}foo{{ else }}bar{{ endif }}'
    self.assertEqual(self.parse(template, var=True), 'foo')
    self.assertEqual(self.parse(template, var=False), 'bar')

  def testElif(self):
    """{{ if }} Elif blocks will be parsed until one matches"""
    template = """
        {{ if [var] == 1 }}a
        {{ elif [var] == 2 }}b
        {{ elif [var] == 3 }}c
        {{ elif [var] == 4 }}d
        {{ endif }}"""
    self.assertEqual(self.parse(template, var=1).strip(), 'a')
    self.assertEqual(self.parse(template, var=2).strip(), 'b')
    self.assertEqual(self.parse(template, var=3).strip(), 'c')
    self.assertEqual(self.parse(template, var=4).strip(), 'd')
    self.assertFalse(self.parse(template, var=5).strip())

  def testIfElifElse(self):
    """{{ if }} Full if/elif/else branch is functional all work"""
    template = """
        {{ if [var] == "a" }}1
        {{ elif [var] == "b"}}2
        {{ else }}3 {{ endif }}"""
    self.assertEqual(self.parse(template, var='a').strip(), '1')
    self.assertEqual(self.parse(template, var='b').strip(), '2')
    self.assertEqual(self.parse(template, var='c').strip(), '3')

  def testSyntaxErrorNoEndif(self):
    """{{ if }} Conditional without {{ endif }} raises TemplateSyntaxError"""
    template = '{{ if [var] }} foo'
    self.assertRaises(templateparser.TemplateSyntaxError, self.parse, template)

  def testSyntaxErrorElifAfterElse(self):
    """{{ if }} An `elif` clause following `else` raises TemplateSyntaxError"""
    template = '{{ if [var] }} {{ else }} {{ elif [var] }} {{ endif }}'
    self.assertRaises(templateparser.TemplateSyntaxError, self.parse, template)

  def testSyntaxErrorDoubleElse(self):
    """{{ if }} Starting a second `else` clause raises TemplateSyntaxError"""
    template = '{{ if [var] }} {{ else }} {{ else }} {{ endif }}'
    self.assertRaises(templateparser.TemplateSyntaxError, self.parse, template)

  def testSyntaxErrorClauseWithoutIf(self):
    """{{ if }} elif / else / endif without `if` raises TemplateSyntaxError"""
    template = '{{ elif }}'
    self.assertRaises(templateparser.TemplateSyntaxError, self.parse, template)
    template = '{{ else }}'
    self.assertRaises(templateparser.TemplateSyntaxError, self.parse, template)
    template = '{{ endif }}'
    self.assertRaises(templateparser.TemplateSyntaxError, self.parse, template)

  def testTagPresence(self):
    """{{ if }} Clauses require the tag to be present as a replacement"""
    template = '{{ if [absent] }} {{ endif }}'
    self.assertRaises(templateparser.TemplateNameError, self.parse, template)

  def testVariableMustBeTag(self):
    """{{ if }} Clauses must reference variables using a tag, not a name"""
    good_template = '{{ if [var] }} x {{ else }} x {{ endif }}'
    self.assertTrue(self.parse(good_template, var='foo'))
    bad_template = '{{ if var }} x {{ else }} x {{ endif }}'
    self.assertRaises(templateparser.TemplateNameError,
                      self.parse, bad_template, var='foo')

  def testLazyEvaluation(self):
    """{{ if }} Variables are retrieved in lazy fashion, not before needed"""
    # Tags are looked up lazily
    template = '{{ if [present] or [absent] }}~{{ endif }}'
    self.assertEqual(self.parse(template, present=True), '~')

    # Indices are looked up lazily
    template = '{{ if [var:present] or [var:absent] }}~{{ endif }}'
    self.assertEqual(self.parse(template, var={'present': 1}), '~')


class TemplateLoops(unittest.TestCase):
  """TemplateParser properly handles for-loops."""
  def setUp(self):
    """Sets up a testbed."""
    self.parser = templateparser.Parser()
    self.parse = self.parser.ParseString
    self.tmpl = templateparser.Template

  def testLoopCount(self):
    """{{ for }} Parser will loop once for each item in the for loop"""
    template = '{{ for num in [values] }}x{{ endfor }}'
    result = self.parse(template, values=range(5))
    self.assertEqual(result, 'xxxxx')

  def testLoopReplaceBasic(self):
    """{{ for }} The loop variable is available via tagname"""
    template = '{{ for num in [values] }}[num],{{ endfor }}'
    result = self.parse(template, values=range(5))
    self.assertEqual(result, '0,1,2,3,4,')

  def testLoopReplaceScope(self):
    """{{ for }} The loop variable overwrites similar names from outer scope"""
    template = '[num], {{ for num in [numbers] }}[num], {{ endfor }}[num]'
    result = self.parse(template, numbers=range(5), num='OUTER')
    self.assertEqual(result, 'OUTER, 0, 1, 2, 3, 4, OUTER')

  def testLoopOverIndexedTag(self):
    """{{ for }} Loops can be performed over indexed tags"""
    template = '{{ for num in [numbers:1] }}x{{ endfor }}'
    result = self.parse( template, numbers=[range(10), range(5), range(10)])
    self.assertEqual(result, 'xxxxx')

  def testLoopVariableIndex(self):
    """{{ for }} Loops variable tags support indexing and functions"""
    template = '{{ for bundle in [bundles]}}[bundle:1:name|upper], {{ endfor }}'
    bundles = [('1', {'name': 'Spam'}), ('2', {'name': 'Eggs'})]
    result = 'SPAM, EGGS, '
    self.parser.RegisterFunction('upper', str.upper)
    self.assertEqual(self.parse(template, bundles=bundles), result)

  def testLoopOnFunctions(self):
    """{{ for }} Loops work on function results if functions are used"""
    template = ('{{ for item in [mapping|items|sorted] }}'
                '[item:0]=[item:1] {{ endfor }}')
    mapping = {'first': 12, 'second': 42}
    result = 'first=12 second=42 '
    self.assertEqual(self.parse(template, mapping=mapping), result)
    # Assert that without sorted, this actually fails
    unsorted = ('{{ for item in [mapping|items] }} '
                '[item:0]=[item:1]{{ endfor }}')
    self.assertNotEqual(self.parse(unsorted, mapping=mapping), result)

  def testLoopTupleAssignment(self):
    """{{ for }} Loops support tuple unpacking for iterators"""
    template = ('{{ for key,val in [mapping|items|sorted] }}'
                '[key]=[val] {{ endfor }}')
    mapping = {'first': 12, 'second': 42}
    result = 'first=12 second=42 '
    self.assertEqual(self.parse(template, mapping=mapping), result)

  def testLoopTupleAssignmentMismatch(self):
    """{{ for }} Loops raise TemplateValueError when tuple unpacking fails"""
    template = '{{ for a, b, c in [iterator] }}[a]{{ endfor }}'
    self.assertEqual(self.parse(template, iterator=['xyz']), 'x')
    self.assertRaises(templateparser.TemplateValueError,
                      self.parse, template, iterator=['eggs'])
    self.assertRaises(templateparser.TemplateValueError,
                      self.parse, template, iterator=range(10))

  def testLoopTagPresence(self):
    """{{ for }} Loops require the loop tag to be present"""
    template = '{{ for item in [absent] }} hello {{ endfor }}'
    self.assertRaises(templateparser.TemplateNameError, self.parse, template)

  def testLoopAbsentIndex(self):
    """{{ for }} Loops over an absent index result in no loops (no error)"""
    template = '{{ for item in [tag:absent] }} x {{ endfor }}'
    self.assertFalse(self.parse(template, tag='absent'))


class TemplateTagPresenceCheck(unittest.TestCase):
  """Test cases for the `ifpresent` TemplateParser construct."""
  def setUp(self):
    self.parser = templateparser.Parser()
    self.parse = self.parser.ParseString
    self.templatefilename = 'ifpresent.utp'

  def testBasicTagPresence(self):
    """{{ ifpresent }} runs the code block if the tag is present"""
    template = '{{ ifpresent [tag] }} hello {{ endif }}'
    self.assertEqual(self.parse(template, tag='spam'), ' hello ')

  def testBasicTagAbsence(self):
    """{{ ifpresent }} does not run the main block if the tag is missing"""
    template = '{{ ifpresent [tag] }} hello {{ endif }}'
    self.assertFalse(self.parse(template))

  def testBasicTagNotPresence(self):
    """{{ ifnotpresent }} runs the code block if the tag is present"""
    template = '{{ ifnotpresent [tag] }} hello {{ endif }}'
    self.assertEqual(self.parse(template, othertag='spam'), ' hello ')

  def testNestedNotPresence(self):
    """{{ ifnotpresent }} runs the code block if the tag is present"""
    template = """{{ ifnotpresent [tag] }}
      {{ ifnotpresent [nestedtag] }}
        hello
      {{ endif }}
    {{ endif }}"""
    self.assertEqual(self.parse(template, othertag='spam').strip(), 'hello')

  def testTagPresenceElse(self):
    """{{ ifpresent }} has a functioning `else` clause"""
    template = '{{ ifpresent [tag] }} yes {{ else }} no {{ endif }}'
    self.assertEqual(self.parse(template, tag='spam'), ' yes ')
    self.assertEqual(self.parse(template), ' no ')

  def testPresenceElif(self):
    """{{ ifpresent }} has functioning `elif` support"""
    template = ('{{ ifpresent [one] }} first '
                '{{ elif [two] }} second {{ else }} third {{ endif }}')
    self.assertEqual(self.parse(template, one='present'), ' first ')
    self.assertEqual(self.parse(template, two='ready'), ' second ', )
    self.assertEqual(self.parse(template), ' third ')

  def testPresenceOfKey(self):
    """{{ ifpresent }} also works on index selectors"""
    template = '{{ ifpresent [tag:6] }} yes {{ else }} no {{ endif }}'
    self.assertEqual(self.parse(template, tag='longtext'), ' yes ')
    self.assertEqual(self.parse(template, tag='short'), ' no ')
    self.assertEqual(self.parse(template), ' no ')

  def testMultiTagPresence(self):
    """{{ ifpresent }} checks the presence of *all* provided tagnames/indices"""
    template = '{{ ifpresent [one] [two] }} good {{ endif }}'
    self.assertEqual(self.parse(template, one=1, two=2), ' good ')
    self.assertFalse(self.parse(template, one=1))
    self.assertFalse(self.parse(template, two=2))

  def testBadSyntax(self):
    """{{ ifpresent }} requires proper tags to be checked for presence"""
    template = '{{ ifpresent var }} {{ endif }}'
    self.assertRaises(templateparser.TemplateSyntaxError, self.parse, template)

  def testMultiTagPresenceFile(self):
    """{{ ifpresent }} checks if multiple runs on a file template containing an
    Ifpresent block work"""

    template = '{{ ifpresent [one] }} [one] {{ endif }}Blank'
    with open(self.templatefilename, 'w') as templatefile:
      templatefile.write(template)
    self.assertEqual(self.parser.Parse(self.templatefilename), 'Blank')
    #self.assertEqual(self.parser.Parse(self.templatefilename), 'Blank')
    #self.assertEqual(self.parser.Parse(self.templatefilename, one=1), ' 1 Blank')

  def tearDown(self):
    for tmpfile in (self.templatefilename, ):
      if os.path.exists(tmpfile):
        if os.path.isdir(tmpfile):
          os.rmdir(tmpfile)
        else:
          os.unlink(tmpfile)


class TemplateStringRepresentations(unittest.TestCase):
  """Test cases for string representation of various TemplateParser parts."""
  def setUp(self):
    self.strip = lambda string: re.sub('\s', '', string)
    self.tmpl = templateparser.Template
    self.parser = templateparser.Parser()

  def testTemplateTag(self):
    """[Representation] TemplateTags str() echoes its literal"""
    template = '[greeting] [title|casing] [person:name|casing] har'
    self.assertEqual(self.strip(str(self.tmpl(template))), self.strip(template))

  def testTemplateConditional(self):
    """[Representation] TemplateConditional str() echoes its literal"""
    template = '{{ if [a] == "foo" }} foo [b] {{ else }} bar [b] {{ endif }}'
    self.assertEqual(self.strip(str(self.tmpl(template))), self.strip(template))

  def testTemplateInline(self):
    """[Representation] TemplateInline str() shows the inlined template part"""
    example = 'Hello [location]'
    template = '{{ inline example }}'
    self.parser['example'] = self.tmpl(example)
    self.assertEqual(self.strip(str(self.tmpl(template, parser=self.parser))),
                     self.strip(example))

  def testTemplateLoop(self):
    """[Representation] TemplateLoop str() echoes its definition"""
    template = ('{{ for a, b in [iter|items] }}{{ for c in [a] }}[c]'
                '{{ endfor }}{{ endfor }}')
    self.assertEqual(self.strip(str(self.tmpl(template))), self.strip(template))


class TemplateNestedScopes(unittest.TestCase):
  """Test cases for nested function scopes."""
  def setUp(self):
    """Sets up a testbed."""
    self.parser = templateparser.Parser()
    self.parse = self.parser.ParseString
    self.tmpl = templateparser.Template

  def testLoopWithInline(self):
    """{{ nested }} Loops can contain an {{ inline }} section"""
    inline = '<li>Hello [name]</li>'
    self.parser['name'] = self.tmpl(inline)
    template = '{{ for name in [names] }}{{ inline name }}{{ endfor }}'
    result = self.parse(template, names=('John', 'Eric'))
    self.assertEqual(result, '<li>Hello John</li><li>Hello Eric</li>')

  def testLoopWithInlineLoop(self):
    """{{ nested }} Loops can contain {{ inline }} loops"""
    inline = '{{ for char in [name] }}[char].{{ endfor }}'
    self.parser['name'] = self.tmpl(inline)
    template = '{{ for name in [names] }}<li>{{ inline name }}</li>{{ endfor }}'
    result = self.parse(template, names=('John', 'Eric'))
    self.assertEqual(result, '<li>J.o.h.n.</li><li>E.r.i.c.</li>')

  def testInlineLoopsInConditional(self):
    """{{ nested }} Inlined loop in a conditional without problems"""
    self.parser['loop'] = self.tmpl('{{ for i in [loops] }}[i]{{ endfor }}')
    self.parser['once'] = self.tmpl('value: [value]')
    tmpl = '{{ if [x] }}{{ inline loop }}{{ else }}{{ inline once }}{{ endif }}'
    result_loop = self.parse(tmpl, loops=range(1, 6), x=True)
    result_once = self.parse(tmpl, value='foo', x=False)
    self.assertEqual(result_loop, '12345')
    self.assertEqual(result_once, 'value: foo')


class TemplateReloading(unittest.TestCase):
  """Tests for FileTemplate automatic reloading upon modification."""
  def setUp(self):
    self.simple = 'simple.html'
    self.simple_raw = 'simple [noun]'
    self.loop = 'loop.html'
    self.loop_raw = '{{ for bit in [blob] }}{{ inline simple.html }}{{ endfor }}'
    with open(self.simple, 'w') as simple:
      simple.write(self.simple_raw)
    with open(self.loop, 'w') as loop:
      loop.write(self.loop_raw)
    self.parser = templateparser.Parser()
    self.parser.AddTemplate(self.simple)
    self.parser.AddTemplate(self.loop)

  def tearDown(self):
    for tmpfile in (self.loop, self.simple):
      if os.path.exists(tmpfile):
        if os.path.isdir(tmpfile):
          os.rmdir(tmpfile)
        else:
          os.unlink(tmpfile)

  def testFileBasicReload(self):
    """[Reload] Template file is reloaded from disk after updating"""
    first = self.parser[self.simple].Parse()
    self.assertEqual(first, self.simple_raw)
    with open(self.simple, 'w') as new_template:
      new_template.write('new content')
      time.sleep(.01) # short pause so that mtime will actually be different
    second = self.parser[self.simple].Parse()
    self.assertEqual(second, 'new content')

  def testInlineReload(self):
    """[Reload] Inlined templates are not automatically reloaded"""
    first = self.parser[self.loop].Parse(blob='four')
    self.assertEqual(first, self.simple_raw * 4)
    with open(self.simple, 'w') as new_template:
      new_template.write('new content')
      time.sleep(.01) # short pause so that mtime will actually be different
    second = self.parser[self.loop].Parse(blob='four')
    self.assertEqual(second, 'new content' * 4)

  def testReloadDeletedTemplate(self):
    """[Reload] Deleted templates are not reloaded and don't trigger errors"""
    os.unlink(self.simple)
    self.assertEqual(self.parser[self.simple].Parse(), self.simple_raw)

  def testReplaceTemplateWithDirectory(self):
    """[Reload] Deleted templates are not reloaded and don't trigger errors"""
    os.unlink(self.simple)
    time.sleep(.01) # short pause so that mtime will actually be different
    os.mkdir(self.simple)
    self.assertEqual(self.parser[self.simple].Parse(), self.simple_raw)


if __name__ == '__main__':
  unittest.main(testRunner=unittest.TextTestRunner(verbosity=2))



#!/usr/bin/python3
"""uWeb3 response classes."""

# Standard modules
try:
  import httplib
except ImportError:
  import http.client as httplib

class Response(object):
  """Defines a full HTTP response.

  The full response consists of a required content part, and then optional
  http response code, cookies, additional headers, and a content-type.
  """
  # Default content-type for Page objects
  CONTENT_TYPE = 'text/html'

  def __init__(self, content='', content_type=CONTENT_TYPE,
               httpcode=200, headers=None, **kwds):
    """Initializes a Response object.

    Arguments:
      @ content: str
        The content to return to the client. This can be either plain text, html
        or the contents of a file (images for example).
      % content_type: str ~~ CONTENT_TYPE ('text/html' by default)
        The Content-Type of the response. This should NOT be set in headers.
      % httpcode: int ~~ 200
        The HTTP response code to attach to the response.
      % headers: dict ~~ None
        A dictionary with header names and their associated values.
    """
    self.charset = kwds.get('charset', 'utf-8')
    self.content = content
    self.httpcode = httpcode
    self.headers = headers or {}
    if (';' not in content_type and
        (content_type.startswith('text/') or
         content_type.startswith('application/json'))):
      content_type = '{!s}; charset={!s}'.format(content_type, self.charset)
    self.content_type = content_type

  # Get and set content-type header
  @property
  def content_type(self):
    """Returns the current Content-Type or None if not set"""
    return self.headers.get('Content-Type', None)

  @content_type.setter
  def content_type(self, content_type):
    """Sets the Content-Type of the response

    Arguments:
      @ content_type: str ~~ CONTENT_TYPE
        The content type of the response.
    """
    current = self.headers.get('Content-Type', '')
    if ';' in current:
      content_type = '{!s}; {!s}'.format(content_type,
                                         current.split(';', 1)[-1])
    self.headers['Content-Type'] = content_type

  def clean_content_type(self):
    """Returns the Content-Type, cleaned from any characters set information."""
    if ';' not in self.headers['Content-Type']:
      return self.headers['Content-Type']
    return self.headers['Content-Type'].split(';')[0]

  # Get and set body text
  @property
  def text(self):
    """Returns the content of this response"""
    return self.content

  @text.setter
  def text(self, content):
    """Sets the content of this response.

    Arguments:
      @ content: str
        The content to return to the client. This can be either plain text, html
        or the contents of a file (images for example).
    """
    self.content = content

  # Retrieve a header list
  @property
  def headerlist(self):
    """Returns the current headers as a list of tuples

    each tuple contains the header key, and its value.
    """
    tuple_list = []
    for key, val in self.headers.items():
      if key == 'Set-Cookie':
        for cookie in val:
          tuple_list.append(
              (key, cookie.encode('ascii', 'ignore').decode('ascii'))
          )
        continue
      if not isinstance(val, str):
        val = str(val)
      tuple_list.append((key, val.encode('ascii', 'ignore').decode('ascii')))
    return tuple_list

  @property
  def status(self):
    """Returns the current http status code for this response."""
    if not self.httpcode:
      return '%d %s' % (500, httplib.responses[500])
    return '%d %s' % (self.httpcode, httplib.responses[self.httpcode])

  def __repr__(self):
    return '<%s instance at %#x>' % (self.__class__.__name__, id(self))

  def __str__(self):
    return self.content

  def SetHeaders(self, headers):
    """Instantly set all headers for this Response """
    self.headers = headers

  def AddHeader(self, header, value):
    """Adds a header to this response's output list"""
    self.headers[header] = value


class Redirect(Response):
  """A response tailored to do redirects."""
  REDIRECT_PAGE = ('<!DOCTYPE html><html><head><title>Page moved</title></head>'
                   '<body>Page moved, please follow <a href="%s">this link</a>'
                   '</body></html>')
  def __init__(self, location, httpcode=307):
    super(Redirect, self).__init__(
        content=self.REDIRECT_PAGE % location,
        content_type='text/html',
        httpcode=httpcode,
        headers={'Location': location})



import os
import sys

import socketio
import eventlet

from uweb3 import uWeb, HotReload
from uweb3.helpers import StaticMiddleware


class SocketMiddleWare(socketio.WSGIApp):
  def __init__(self, socketio_server, uweb3_server, socketio_path='socket.io'):
    super(SocketMiddleWare, self).__init__(socketio_server,
                                           uweb3_server,
                                           socketio_path=socketio_path)

class Uweb3SocketIO:
  def __init__(self, app, sio, static_dir=None):
    if not isinstance(app, uWeb):
      raise Exception("App must be an uWeb3 instance!")
    if not static_dir:
      static_dir = os.path.dirname(os.path.abspath(__file__))

    self.host = app.config.options['development'].get('host', '127.0.0.1')
    self.port = app.config.options['development'].get('port', 8000)
    if app.config.options['development'].get('dev', False) == 'True':
      hotreload = app.config.options['development'].get('uweb_dev', 'False')
      HotReload(app.executing_path, uweb_dev=hotreload)
    self.setup_app(app, sio, static_dir)


  def setup_app(self, app, sio, static_dir):
    path = os.path.join(app.executing_path, 'static')
    app = SocketMiddleWare(sio, app)
    static_directory = [os.path.join(sys.path[0], path)]
    app = StaticMiddleware(app, static_root='static',
                           static_dirs=static_directory)
    eventlet.wsgi.server(eventlet.listen((self.host, int(self.port))), app)



#!/usr/bin/python3
"""uWeb3 model base classes."""

# Standard modules
import os
import datetime
import sys
import hashlib
import pickle
import secrets
import configparser

from contextlib import contextmanager


class Error(Exception):
  """Superclass used for inheritance and external exception handling."""

class DatabaseError(Error):
  """Superclass for errors returned by the database backend."""

class CurrentlyWorking(Error):
  """Caching error"""

class BadFieldError(DatabaseError):
  """A field in the record could not be written to the database."""

class AlreadyExistError(Error):
  """The resource already exists, and cannot be created twice."""

class NotExistError(Error):
  """The requested or provided resource doesn't exist or isn't accessible."""

class PermissionError(Error):
  """The entity has insufficient rights to access the resource."""


class SettingsManager(object):
  def __init__(self, filename=None, path=None):
    """Creates a ini file with the child class name

    Arguments:
      % filename: str, optional
      Name of the file, optionally without the extension will default to .ini
      If not filename is given the class.__name__ will be used to look for the config file in the path
      % path: str, Optional
      Path to the config file, will be used if filename is relative, eg does not start with '/'
    """
    self.options = None
    extension = '' if filename and filename.endswith(('.ini', '.conf')) else '.ini'

    if filename:
      self.filename = f"{filename[:1].lower() + filename[1:] + extension}"
    else:
      self.filename = self.TableName() + extension

    if path and not filename.startswith('/'):
      self.file_location = os.path.join(path, self.filename)
    else:
      self.file_location = self.filename
    self.__CheckPermissions()
    if not os.path.isfile(self.file_location):
      os.mknod(self.file_location)

    self.mtime = None
    self.config = configparser.ConfigParser()
    self.Read()

  @classmethod
  def TableName(cls):
    """Returns the 'database' table name for the SettingsManager class.

    If this is not explicitly defined by the class constant `_TABLE`, the return
    value will be the class name with the first letter lowercased.
    We stick to the same naming scheme as for more table like connectors even
    though we use files instead of tables in this class.
    """
    if cls._TABLE:
      return cls._TABLE
    name = cls.__name__
    return name[0].lower() + name[1:]

  def __CheckPermissions(self):
    """Checks if SettingsManager can read/write to file."""
    if not os.path.isfile(self.file_location):
      return True
    if not os.access(self.file_location, os.R_OK):
      raise PermissionError(f"SettingsManager missing permissions to read file: {self.file_location}")
    if not os.access(self.file_location, os.W_OK):
      raise PermissionError(f"SettingsManager missing permissions to write to file: {self.file_location}")

  def Create(self, section, key, value):
    """Creates a section or/and key = value

    Arguments:
      @ section: str
        Name of the section you want to create or append key = value to
      @ key: str
        Name of the key you want to create
      @ value: str

    Raises:
      ValueError
    """
    if not self.options.get(section):
      self.config.add_section(section)
    else:
      if self.config[section].get(key):
        raise ValueError("key already exists")

    self.config.set(section, key, value)
    self._Write(False)
    self.mtime = None

  def Read(self):
    """Reads the config file and populates the options member
    It uses the mtime to see if any re-reading is required"""
    if not self.mtime:
      curtime = os.path.getmtime(self.file_location)
      if self.mtime and self.mtime == curtime:
        return False
      self.config.read(self.file_location)
      self.options = self.config._sections
      self.mtime = curtime
    return True

  def Update(self, section, key, value):
    """Updates ini file
    After update reads file again and updates options attribute

    Arguments:
      @ section: str
      @ key: str
      @ value: str

    Raises
      TypeError: Option values must be string
    """
    if not self.options.get(section):
      self.config.add_section(section)
    self.config.set(section, key, value)
    self._Write()
    self.mtime = None

  def Delete(self, section, key=None):
    """Delete sections/keys from the INI file
    Be aware, deleting a section that is not empty will remove all keys from that
    given section

    Arguments:
      @ section: str
        Name of the section
      @ key: None / str
        Name of the key you want to remove
        If set to None (default) it will delete the supplied section
    Raises:
      configparser.NoSectionError
    """
    if key:
      self.config.remove_option(section, key)
    if not key:
      self.config.remove_section(section)
    self._Write()
    self.mtime = None
    return True

  def _Write(self, reread=True):
    """Internal function to store the current config to file"""
    with open(self.file_location, 'w') as configfile:
      self.config.write(configfile)
    if reread:
      return self.Read()
    return True


class SecureCookie(object):
  """The secureCookie class works just like other data abstraction classes,
  except that it stores its data in client side cookies that are signed with a
  server side secret to avoid tampering by the end-user.

  Subclass this class with your own class to create signed cookie objects. The
  name for your class will reflect the name for the cookie in the cookie.
  """

  HASHTYPE = 'ripemd160'
  _TABLE = None
  _CONNECTOR = 'signedCookie'

  def __init__(self, connection):
    """Create a new SecureCookie instance."""
    self.connection = connection
    self.request, self.cookies, self.cookie_salt = self.connection
    self.debug = self.connection.debug
    self._rawcookie = None
    if self.debug:
      print('current cookies (unvalidated) for request:', self.cookies)

  def __str__(self):
    """Returns the cookie's value if it was valid and untampered with."""
    return str(self.rawcookie)

  @classmethod
  def TableName(cls):
    """Returns the 'database' table name for the SecureCookie class.

    If this is not explicitly defined by the class constant `_TABLE`, the return
    value will be the class name with the first letter lowercased.
    We stick to the same naming scheme as for more table like connectors even
    though we use cookies instead of tables in this class.
    """
    if cls._TABLE:
      return cls._TABLE
    name = cls.__name__
    return name[0].lower() + name[1:]

  @property
  def rawcookie(self):
    """Reads the request cookie, checks if it was signed correctly and return
    the value, or returns False"""
    if not self._rawcookie is None:
      return self._rawcookie
    name = self.TableName()
    if name in self.cookies and self.cookies[name]:
      isValid, value = self.__ValidateCookieHash(self.cookies[name])
      if isValid:
        self._rawcookie = value
        return value
      if self.debug:
        print('Secure cookie "%s" was tampered with and thus invalid. content was: %s ' % (name, self.cookies[name]))
    if self.debug:
      print('Secure cookie "%s" was not present.' % name)
    self._rawcookie = ''
    return self._rawcookie

  @classmethod
  def Create(cls, connection, data, **attrs):
    """Creates a secure cookie

    Arguments:
      @ data: dict
        Needs to have a key called __name with value of how you want to name the 'table'
      % only_return_hash: boolean
        If this is set it will just return the hash of the cookie. This is used to
        validate the cookies hash
      % update: boolean
        Used to update the cookie. Updating actually means deleting and setting a new
        one. This attribute is used by the update method from this class
      % expires: str ~~ None
        The date + time when the cookie should expire. The format should be:
        "Wdy, DD-Mon-YYYY HH:MM:SS GMT" and the time specified in UTC.
        The default means the cookie never expires.
        N.B. Specifying both this and `max_age` leads to undefined behavior.
      % path: str ~~ '/'
        The path for which this cookie is valid. This default ('/') is different
        from the rule stated on Wikipedia: "If not specified, they default to
        the domain and path of the object that was requested".
      % domain: str ~~ None
        The domain for which the cookie is valid. The default is that of the
        requested domain.
      % max_age: int
        The number of seconds this cookie should be used for. After this period,
        the cookie should be deleted by the client.
        N.B. Specifying both this and `expires` leads to undefined behavior.
      % secure: boolean
        When True, the cookie is only used on https connections.
      % httponly: boolean
        When True, the cookie is only used for http(s) requests, and is not
        accessible through Javascript (DOM).

    Raises:
      ValueError: When cookie with name already exists
    """
    cls.connection = connection
    cls.request, cls.cookies, cls.cookie_salt = connection
    name = cls.TableName()
    cls._rawcookie = data

    hashed = cls.__CreateCookieHash(cls, data)
    cls.cookies[name] = hashed
    cls.request.AddCookie(name, hashed, **attrs)
    return cls

  def Update(self, data, **attrs):
    """"Updates a secure cookie
    Keep in mind that the actual cookie's value is avilable from the next
    request. After calling this method it will update the cookie attribute to
    the new value however.

    Arguments:
      @ data: dict
        Needs to have a key called __name with value of how you want to name the 'table'
      % only_return_hash: boolean
        If this is set it will just return the hash of the cookie. This is used to
        validate the cookies hash
      % update: boolean
        Used to update the cookie. Updating actually means deleting and setting a new
        one. This attribute is used by the update method from this class
      % expires: str ~~ None
        The date + time when the cookie should expire. The format should be:
        "Wdy, DD-Mon-YYYY HH:MM:SS GMT" and the time specified in UTC.
        The default means the cookie never expires.
        N.B. Specifying both this and `max_age` leads to undefined behavior.
      % path: str ~~ '/'
        The path for which this cookie is valid. This default ('/') is different
        from the rule stated on Wikipedia: "If not specified, they default to
        the domain and path of the object that was requested".
      % domain: str ~~ None
        The domain for which the cookie is valid. The default is that of the
        requested domain.
      % max_age: int
        The number of seconds this cookie should be used for. After this period,
        the cookie should be deleted by the client.
        N.B. Specifying both this and `expires` leads to undefined behavior.
      % secure: boolean
        When True, the cookie is only used on https connections.
      % httponly: boolean
        When True, the cookie is only used for http(s) requests, and is not
        accessible through Javascript (DOM).

    Raises:
      ValueError: When no cookie with given name found
    """
    name = self.TableName()
    if not self.rawcookie:
      raise ValueError("No valid cookie with name `{}` found".format(name))
    self._rawcookie = data
    self.request.AddCookie(name,  self.__CreateCookieHash(data), **attrs)

  def Delete(self):
    """Deletes cookie based on name
    The cookie is no longer in the session after calling this method
    """
    name = self.TableName()
    self.request.DeleteCookie(name)
    self._rawcookie = None

  def __CreateCookieHash(self, data):
    hex_string = pickle.dumps(data).hex()

    hashed = (hex_string + self.cookie_salt).encode('utf-8')
    h = hashlib.new(self.HASHTYPE)
    h.update(hashed)
    return '{}+{}'.format(h.hexdigest(), hex_string)

  def __ValidateCookieHash(self, cookie):
    """Takes a cookie and validates it

    Arguments:
      @ str: A hashed cookie from the `__CreateCookieHash` method
    """
    if not cookie:
      return None
    try:
      data = cookie.rsplit('+', 1)[1]
      data = pickle.loads(bytes.fromhex(data))
    except Exception:
      return (False, None)

    if cookie == self.__CreateCookieHash(data):
      return (True, data)
    return (False, None)


# Record classes have many methods, this is not an actual problem.
# pylint: disable=R0904
class BaseRecord(dict):
  """Basic database record wrapping class.

  This allows structured database manipulation for applications. Supported
  features include:
  * Loading a record from Primary Key;
  * Deleting a record by Primary Key;
  * Deleting an existing open record;
  * Listing all records of the current type;
  * Calculating the minimum changed set and storing this to the database.
  """
  _LOAD_METHOD = 'FromPrimary'
  _PRIMARY_KEY = 'ID'
  _TABLE = None

  def __init__(self, connection, record, run_init_hook=True):
    """Initializes a BaseRecord instance.

    Arguments:
      @ connection: object
        The database connection to use for further queries.
      @ record: mapping
        A field:value mapping of the database record information.
      % run_init_hook: bool ~~ True
        States whether or not the `_PostInit()` hook should be run after
        the other initialization steps have completed.
    """
    super(BaseRecord, self).__init__(record)
    if not hasattr(BaseRecord, '_SUBTYPES'):
      # Adding classes at runtime is pretty rare, but fails this code.
      BaseRecord._SUBTYPES = dict(RecordTableNames())
    self.connection = connection
    self._record = self._DataRecord()
    # _PostInit hook should run after making a live copy of the data, so that
    # mirrored data transforms between _PostInit and _PreSave will not trigger
    # the record to be updated on saves where the data hasn't actually changed.
    if run_init_hook:
      self._PostInit()

  def __eq__(self, other):
    """Simple equality comparison for database objects.

    To compare equal, two objects must:
      1) Be of the same type;
      2) Have the same primary key which is NOT None;
      3) Have the same content.

    In the case that the compared objects have foreign relations, these  will be
    compared as well (recursively). If only one of the objects has foreign
    relations loaded, only the primary key value will be compared to the value
    in the other Record.
    """
    if type(self) != type(other):
      return False  # Types must be the same.
    elif not (self.key == other.key is not None):
      return False  # Records should have the same non-None primary key value.
    elif len(self) != len(other):
      return False  # Records must contain the same number of objects.
    for key, value in self.items():
      other_value = other[key]
      if isinstance(value, BaseRecord) != isinstance(other_value, BaseRecord):
        # Only one of the two is a BaseRecord instance
        if (isinstance(value, BaseRecord) and value.key != other_value or
            isinstance(other_value, BaseRecord) and other_value.key != value):
          return False
      elif value != other_value:
        return False
    return True

  def __hash__(self):
    """Returns the hashed value of the key."""
    return hash(self.key)

  def __int__(self):
    """Returns the integer key value of the Record.

    For record objects where the primary key value is not (always) an integer,
    this function will raise an error in the situations where it is not.
    """
    key_val = self._ValueOrPrimary(self.key)
    if not isinstance(key_val, (int)):
      # We should not truncate floating point numbers.
      # Nor turn strings of numbers into an integer.
      raise ValueError('The primary key is not an integral number.')
    return key_val

  def __ne__(self, other):
    """Returns the proper inverse of __eq__."""
    # Without this, the non-equal checks used in __eq__ will not work,
    # and the  `!=` operator would not be the logical inverse of `==`.
    return not self == other

  def __repr__(self):
    return '%s(%s)' % (type(self).__name__, super(BaseRecord, self).__repr__())

  def __str__(self):
    return '%s({%s})' % (
        self.__class__.__name__,
        ', '.join('%r: %r' % item for item in self.items()))

  def copy(self):
    """Returns a shallow copy of the Record that is a new functional Record."""
    return self.__class__(
        self.connection, super(BaseRecord, self).copy(), run_init_hook=False)

  # ############################################################################
  # Rich comparators
  #
  def __gt__(self, other):
    """Index of this record is greater than the other record's.

    This requires both records to be of the same record class."""
    if type(self) == type(other):
      return self.key > other.key
    return NotImplemented

  def __ge__(self, other):
    """Index of this record is greater than, or equal to, the other record's.

    This requires both records to be of the same record class."""
    if type(self) == type(other):
      return self.key >= other.key
    return NotImplemented

  def __lt__(self, other):
    """Index of this record is smaller than the other record's.

    This requires both records to be of the same record class."""
    if type(self) == type(other):
      return self.key < other.key
    return NotImplemented

  def __le__(self, other):
    """Index of this record is smaller than, or equal to, the other record's.

    This requires both records to be of the same record class."""
    if type(self) == type(other):
      return self.key <= other.key
    return NotImplemented

  # ############################################################################
  # Hooks for things to run before and after creating and updating records
  #
  def _PreCreate(self, _cursor):
    """Hook that runs before creating (inserting) a Record in the database.

    Typically you would verify values of the Record in this step, or transform
    the data for database-safe insertion. If the data is transformed here, this
    transformation should be reversed in `_PostCreate()`.

    Returning False from this method will halt the creation of the record.
    """
    return True

  def _PreSave(self, _cursor):
    """Hook that runs before saving (updating) a Record in the database.

    Typically you would verify values of the Record in this step, or transform
    the data for database-safe insertion. If the data is transformed here, this
    transformation should be reversed in `_PostSave()`.

    Returning False from this method will halt the updating of the record.
    """
    return True

  def _PostInit(self):
    """Hook that runs after initializing a Record instance.

    This typically runs on every instance, but can be suppressed. Records that
    are newly created using the `Create()` classmethod will NOT have this hook
    run. They should use `_PreCreate()` and `_PostCreate()` methods.

    Any transforms done after creating and saving should likely be present here.
    """

  def _PostCreate(self, _cursor):
    """Hook that runs after creating (inserting) a Record in the database.

    Any transforms that were performed on the data should be reversed here.
    """

  def _PostSave(self, _cursor):
    """Hook that runs after saving (updating) a Record in the database.

    Any transforms that were performed on the data should be reversed here.
    """

  def _PreDelete(self):
    """Hook that runs before deleting a Record in the database.
    Returning False from this method will halt the deletion of the record.
    """
    return True

  def _PostDelete(self):
    """Hook that runs after deleting a Record in the database."""

  # ############################################################################
  # Base record functionality methods, to be implemented by subclasses.
  # Some methods have a generic implementation, but may need customization,
  #
  @classmethod
  def Create(cls, connection, record):
    """Creates a proper record object and stores it to the database.

    After storing it to the database, the live object is returned

    Arguments:
      @ connection: object
        Database connection to use for the created record..
      @ record: mapping
        The record data to write to the database.

    Returns:
      BaseRecord: the record that was created from the initiation mapping.
    """
    raise NotImplementedError

  @classmethod
  def DeletePrimary(cls, connection, key):
    """Deletes a database record based on the primary key value.

    Arguments:
      @ connection: object
        Database connection to use.
      @ pkey_value: obj
        The value for the primary key field
    """
    raise NotImplementedError

  def Delete(self):
    """Deletes a loaded record based on `self.TableName` and `self.key`.

    For deleting an unloaded object, use the classmethod `DeletePrimary`.
    """
    if not self._PreDelete():
      return False
    self.DeletePrimary(self.connection, self.key)
    self._record.clear()
    self.clear()
    self._PostDelete()

  @classmethod
  def FromPrimary(cls, connection, pkey_value):
    """Returns the Record object that belongs to the given primary key value.

    Arguments:
      @ connection: object
        Database connection to use.
      @ pkey_value: obj
        The value for the primary key field

    Raises:
      NotExistError:
        There is no record that matches the given primary key value.

    Returns:
      Record: Database record abstraction class.
    """
    raise NotImplementedError

  @classmethod
  def List(cls, connection, conditions=None):
    """Yields a Record object for every table entry.

    Arguments:
      @ connection: object
        Database connection to use.
      % conditions: str / iterable ~~ None
        Optional query portion that will be used to limit the list of results

    Yields:
      Record: Database record abstraction class.
    """
    raise NotImplementedError

  def Save(self):
    """Saves the changes made to the record.

    This performs an update to the record, except when `create_new` if set to
    True, in which case the record is inserted.

    Arguments:
      % create_new: bool ~~ False
        Tells the method to create a new record instead of updating a current.
        This should be used when Save is called by the Create() method.
    """
    raise NotImplementedError

  @classmethod
  def _LoadAsForeign(cls, connection, relation_value, method=None):
    """Loads a record as a foreign relation of another.

    Defaults to using the _LOAD_METHOD defined on the class, but when
    provided the optional `method` argument, this named method is used instead.
    """
    if method is None:
      method = cls._LOAD_METHOD
    return getattr(cls, method)(connection, relation_value)

  # ############################################################################
  # Functions for tracking table and primary key values
  #
  def _Changes(self):
    """Returns the differences of the current state vs the last stored state."""
    sql_record = self._DataRecord()
    return {
        key: value
        for key, value in sql_record.items() if self._record.get(key) != value
    }

  def _DataRecord(self):
    """Returns a dictionary of the record's database values
    For any Record object present, its primary key value (`Record.key`) is used.
    """
    return {key: self._ValueOrPrimary(value) for key, value in super().items()}

  @staticmethod
  def _ValueOrPrimary(value):
    """Returns the value, or its primary key value if it's a Record."""
    while isinstance(value, BaseRecord):
      if hasattr(value, '_RECORD_KEY') and value._RECORD_KEY:
        value = value[value._RECORD_KEY]
      else:
        value = value.key
    return value

  @classmethod
  def TableName(cls):
    """Returns the database table name for the Record class.

    If this is not explicitly defined by the class constant `_TABLE`, the return
    value will be the class name with the first letter lowercased.
    """
    if cls._TABLE:
      return cls._TABLE
    name = cls.__name__
    return name[0].lower() + name[1:]

  # Pylint falsely believes this property is overwritten by its setter later on.
  # pylint: disable=E0202
  @property
  def key(self):
    """Returns the primary key for the object.

    This is used for the Save/Update methods, where foreign relations should be
    stored by their primary key.
    """
    if isinstance(self._PRIMARY_KEY, tuple):
      record = self._DataRecord()
      return tuple(record[key] for key in self._PRIMARY_KEY)
    return self.get(self._PRIMARY_KEY)
  # pylint: enable=E0202

  # Pylint doesn't understand property setters at all.
  # pylint: disable=E0102, E0202, E1101
  @key.setter
  def key(self, value):
    """Sets the value of the primary key."""
    if isinstance(value, tuple):
      if len(value) != len(self._PRIMARY_KEY):
        raise ValueError('Not enough values for compound key.')
      for key, key_val in zip(self._PRIMARY_KEY, value):
        self[key] = key_val
    else:
      self[self._PRIMARY_KEY] = value
  # pylint: enable=E0102, E0202, E1101

  Error = Error
  AlreadyExistError = AlreadyExistError
  NotExistError = NotExistError
  PermissionError = PermissionError


class Record(BaseRecord):
  """Extensions to the Record abstraction for relational database use."""
  _FOREIGN_RELATIONS = {}
  _CONNECTOR = 'mysql'
  SEARCHABLE_COLUMNS = []

  # ############################################################################
  # Methods enabling auto-loading
  #
  def GetRaw(self, field):
    """Returns the value of the field, suppressing auto-loading."""
    return super(Record, self).__getitem__(field)

  def __getitem__(self, field):
    """Returns the value corresponding to a given `field`.

    If a field represents a foreign relation, this will be delegated to
    the `_LoadForeign` method.
    """
    value = super(Record, self).__getitem__(field)
    return self._LoadForeign(field, value)

  def _LoadForeign(self, field, value):
    """Loads and returns objects referenced by foreign key.

    This is done by checking the `field` against the class' `_FOREIGN_RELATIONS`
    mapping. If a match is found, `_LoadForeignFromRelationsTable` is executed
    and its return value returned.

    If the `field` is not present in the class mapping, it will be checked
    against table names for each of the subclasses of Record. This mapping is
    maintained in `_SUBTYPES`. If a match is found, an instance of the
    corresponding class will replace the existing value, and will subsequently
    be returned.

    If the `field` is not present in either mapping, its value will remain
    unchanged, and returned as such.

    N.B. If the field name the same as the record's `TableName`, it will NOT be
    automatically resolved. The assumption is that the field will not contain a
    meaningful reference. This behavior can be altered by specifying the
    relation in the _FOREIGN_RELATIONS class constant.

    Arguments:
      @ field: str
        The field name to be checked for foreign references
      @ value: obj
        The current value for the field. This is used as lookup index in case
        of foreign references.

    Returns:
      obj: The value belonging to the given `field`. In case of resolved foreign
           references, this will be the referenced object. Else it's unchanged.
    """
    if value is None:
      return None
    elif not isinstance(value, BaseRecord):
      if field in self._FOREIGN_RELATIONS:
         value = self._LoadUsingForeignRelations(
            self._FOREIGN_RELATIONS[field], field, value)
      elif field == self.TableName():
        return value
      elif field in self._SUBTYPES:
        value = self._SUBTYPES[field]._LoadAsForeign(self.connection, value)
      self[field] = value
    return value

  def _LoadUsingForeignRelations(self, foreign_cls, field, value):
    """Loads and returns foreign relation based on given class (name).

    The action taken depends on the given `cls`. If the given class is None (or
    otherwise boolean false), no action will be taken, and the value will be
    returned unchanged.

    If the class is given as string, it will be loaded from the current module.
    It should be a proper subclass of Record, after which the current `value` is
    used to create a record using `cls._LoadAsForeign`.

    Arguments:
      @ foreign_cls: Record / str / dict
        The class name or actual type to create an instance from. Could also
        be a dictionary with `class` and `loader` keys that indicate class and
        method to use for loading foreign relations.
      @ field: str
        The field name to be checked for foreign references
      @ value: obj
        The current value for the field. This is used as lookup index in case
        of foreign references.

    Raises:
      ValueError: If the class name cannot be found, or the type is not a
                  subclass of Record.

    Returns:
      obj: The value belonging to the given `field`. In case of resolved foreign
           references, this will be the referenced object. Else it's unchanged.
    """
    def GetRecordClass(cls):
      """Returns the record class or loads it from its string name"""
      if isinstance(cls, str):
        try:
          cls = getattr(sys.modules[self.__module__], cls)
        except AttributeError:
          raise ValueError(
              'Bad _FOREIGN_RELATIONS map: Target %r not a class in %r' % (
                  cls, self.__module__))
      if not issubclass(cls, Record):
        raise ValueError('Bad _FOREIGN_RELATIONS map: '
                         'Target %r not a subclass of Record' % cls.__name__)
      return cls

    if foreign_cls is None:
      return value
    if type(foreign_cls) is dict:
      cls = GetRecordClass(foreign_cls['class'])
      loader = foreign_cls.get('loader')
      return cls._LoadAsForeign(self.connection, value, method=loader)
    return GetRecordClass(foreign_cls)._LoadAsForeign(self.connection, value)

  # ############################################################################
  # Override basic dict methods so that autoload mechanisms function on them.
  #
  def get(self, key, default=None):
    """Returns the value for `key` if its present, otherwise `default`."""
    try:
      return self[key]
    except KeyError:
      return default

  def pop(self, field, *default):
    """Pops the value corresponding to the field from the Record.

    If the field does not exist, either KeyError or an optional default value
    is returned instead.
    """
    try:
      value = self[field]
    except KeyError:
      if not default:
        raise
      return default[0]
    del self[field]
    return value

  def iteritems(self):
    """Yields all field+value pairs in the Record.

    N.B. This automatically resolves foreign references.
    """
    return ((key, self[key]) for key in self)

  def itervalues(self):
    """Yields all values in the Record, loading foreign references."""
    return (self[key] for key in self)

  def items(self):
    """Returns a list of field+value pairs in the Record.

    N.B. This automatically resolves foreign references.
    """
    return list(self.iteritems())

  def values(self):
    """Returns a list of values in the Record, loading foreign references."""
    return list(self.itervalues())

  # ############################################################################
  # Private methods to be used for development
  #
  @classmethod
  def _FromParent(cls, parent, relation_field=None, conditions=None,
                 limit=None, offset=None, order=None,
                 yield_unlimited_total_first=False):
    """Returns all `cls` objects that are a child of the given parent.

    This utilized the parent's _Children method, with either this class'
    TableName or the filled out `relation_field`.

    Arguments:
      @ parent: Record
        The parent for who children should be found in this class
      % relation_field: str ~~ cls.TableName()
        The fieldname in this class' table which relates to the parent's primary
        key. If not given, parent.TableName() will be used.
      % conditions: str / iterable ~~ None
        The extra condition(s) that should be applied when querying for records.
      % limit: int ~~ None
        Specifies a maximum number of items to be yielded. The limit happens on
        the database side, limiting the query results.
      % offset: int ~~ None
        Specifies the offset at which the yielded items should start. Combined
        with limit this enables proper pagination.
      % order: iterable of str/2-tuple
        Defines the fields on which the output should be ordered. This should be
        a list of strings or 2-tuples. The string or first item indicates the
        field, the second argument defines descending order
        (desc. if True).
      % yield_unlimited_total_first: bool ~~ False
        Instead of yielding only Record objects, the first item returned is the
        number of results from the query if it had been executed without limit.
    """
    if not isinstance(parent, Record):
      raise TypeError('parent argument should be a Record type.')
    relation_field = relation_field or parent.TableName()
    relation_value = parent.connection.EscapeValues(cls._ValueOrPrimary(parent))
    qry_conditions = ['`%s` = %s' % (relation_field, relation_value)]
    if conditions:
      if isinstance(conditions, str):
        qry_conditions.append(conditions)
      else:
        qry_conditions.extend(conditions)

    firstrow = yield_unlimited_total_first # set a flag to skip the linking of
    # the first row to our parent, as that will be the full record cound instead
    # of a record
    for record in cls.List(parent.connection, conditions=qry_conditions,
        limit=limit, offset=offset, order=order,
        yield_unlimited_total_first=yield_unlimited_total_first):
      if not firstrow:
        record[relation_field] = parent.copy()
      firstrow = False
      yield record

  def _Children(self, child_class, relation_field=None, conditions=None,
    limit=None, offset=None, order=None, yield_unlimited_total_first=False):
    """Returns all `child_class` objects related to this record.

    The table for the given `child_class` will be queried for all fields where
    the `relation_field` is the same as this record's primary key (`self.key`).

    These records will then be yielded as instances of the child class.

    Arguments:
      @ child_class: type (Record subclass)
        The child class whose objects should be found.
      % relation_field: str ~~ self.TableName()
        The fieldname in the `child_class` table which relates that table to the
        table for this record.
      % conditions: str / iterable ~~
        The extra condition(s) that should be applied when querying for records.
      % limit: int ~~ None
        Specifies a maximum number of items to be yielded. The limit happens on
        the database side, limiting the query results.
      % offset: int ~~ None
        Specifies the offset at which the yielded items should start. Combined
        with limit this enables proper pagination.
      % order: iterable of str/2-tuple
        Defines the fields on which the output should be ordered. This should be
        a list of strings or 2-tuples. The string or first item indicates the
        field, the second argument defines descending order (desc. if True).
      % yield_unlimited_total_first: bool ~~ False
        Instead of yielding only Record objects, the first item returned is the
        number of results from the query if it had been executed without limit.
    """
    # Delegating to let child class handle its own querying. These are methods
    # for development, and are private only to prevent name collisions.
    # pylint: disable=W0212
    return child_class._FromParent(
        self, relation_field=relation_field, conditions=conditions,
        limit=limit, offset=offset, order=order,
        yield_unlimited_total_first=yield_unlimited_total_first)

  def _DeleteChildren(self, child_class, relation_field=None):
    """Deletes all `child_class` objects related to this record.

    The table for the given `child_class` will be queried for all fields where
    the `relation_field` is the same as this record's primary key (`self.key`).

    Arguments:
      @ child_class: type (Record subclass)
        The child class whose objects should be deleted.
      % relation_field: str ~~ self.TableName()
        The fieldname in the `child_class` table which relates that table to the
        table for this record.
    """
    relation_field = relation_field or self.TableName()
    with self.connection as cursor:
      safe_key = self.connection.EscapeValues(self.key)
      cursor.Delete(table=child_class.TableName(),
                    conditions='`%s`=%s' % (relation_field, safe_key))

  @classmethod
  def _PrimaryKeyCondition(cls, connection, value):
    """Returns the primary key condition to be used."""
    if isinstance(cls._PRIMARY_KEY, tuple):
      if not isinstance(value, tuple):
        raise TypeError(
            'Compound keys should be loaded using a tuple of key values.')
      if len(value) != len(cls._PRIMARY_KEY):
        raise ValueError('Not enough values (%d) for compound key.', len(value))
      values = tuple(map(cls._ValueOrPrimary, value))
      return ' AND '.join('`%s` = %s' % (field, value) for field, value
                   in zip(cls._PRIMARY_KEY, connection.EscapeValues(values)))
    return '`%s`.`%s` = %s' % (
        cls.TableName(),
        cls._PRIMARY_KEY,
        connection.EscapeValues(cls._ValueOrPrimary(value)))

  def _RecordCreate(self, cursor):
    """Inserts the record's current values in the database as a new record.

    Upon success, the record's primary key is set to the result's insertid
    """
    try:
      values = self._DataRecord()
      # Compound key case
      if isinstance(self._PRIMARY_KEY, tuple):
        auto_inc_field = set(self._PRIMARY_KEY) - set(values)
        if auto_inc_field:
          raise ValueError('No value for compound key field(s): %s' % (
              ', '.join(map(repr, auto_inc_field))))
        return cursor.Insert(table=self.TableName(), values=values)
      # Single-column key case
      result = cursor.Insert(table=self.TableName(), values=values)
      if result.insertid:
        self._record[self._PRIMARY_KEY] = self.key = result.insertid
    except cursor.OperationalError as err_obj:
      if err_obj[0] == 1054:
        raise BadFieldError(err_obj[1])
      raise DatabaseError(err_obj)

  def _RecordUpdate(self, cursor):
    """Updates the existing database entry with the record's current values.

    The constraint with which the record is updated is the name and value of the
    Record's primary key (`self._PRIMARY_KEY` and `self.key` resp.)
    """
    try:
      if isinstance(self._PRIMARY_KEY, tuple):
        primary = tuple(self._record[key] for key in self._PRIMARY_KEY)
      else:
        primary = self._record[self._PRIMARY_KEY]
      cursor.Update(
          table=self.TableName(),
          values=self._Changes(),
          conditions=self._PrimaryKeyCondition(self.connection, primary))
    except KeyError:
      raise Error('Cannot update record without pre-existing primary key.')
    except cursor.OperationalError as err_obj:
      if err_obj[0] == 1054:
        raise BadFieldError(err_obj[1])
      raise DatabaseError(err_obj)

  def _SaveForeign(self, cursor):
    """Recursively saves all nested Record instances."""
    for value in super(Record, self).items():
      if isinstance(value, Record):
        # Accessing protected members of a foreign class. Also, the only means
        # of recursively saving the record tree without opening multiple
        # database transactions (which would lead to exceptions really fast).
        # pylint: disable=W0212
        value._SaveForeign(cursor)
        value._SaveSelf(cursor)

  def _SaveSelf(self, cursor):
    """Updates the existing database entry with the record's current values.

    The constraint with which the record is updated is the name and value of the
    Record's primary key (`self._PRIMARY_KEY` and `self.key` resp.)
    """
    self._PreSave(cursor)
    difference = self._Changes()
    if difference:
      self._RecordUpdate(cursor)
      self._record.update(difference)
    self._PostSave(cursor)

  # ############################################################################
  # Public methods for creation, deletion and storing Record objects.
  #
  @classmethod
  def Create(cls, connection, record):
    record = cls(connection, record, run_init_hook=False)
    with connection as cursor:
      # Accessing protected members of a foreign class.
      # pylint: disable=W0212
      record._PreCreate(cursor)
      record._RecordCreate(cursor)
      record._PostCreate(cursor)
    return record

  @classmethod
  def DeletePrimary(cls, connection, pkey_value):
    with connection as cursor:
      cursor.Delete(table=cls.TableName(),
                    conditions=cls._PrimaryKeyCondition(connection, pkey_value))

  @classmethod
  def FromPrimary(cls, connection, pkey_value):
    with connection as cursor:
      record = cursor.Select(
          table=cls.TableName(),
          conditions=cls._PrimaryKeyCondition(connection, pkey_value))
    if not record:
      raise NotExistError('There is no %r for primary key %r' % (
          cls.__name__, pkey_value))
    return cls(connection, record[0])

  @classmethod
  def List(cls, connection, conditions=None, limit=None, offset=None,
           order=None, yield_unlimited_total_first=False, search=None,
           tables=None, escape=True, fields=None, distinct=False):
    """Yields a Record object for every table entry.

    Arguments:
      @ connection: object
        Database connection to use.
      % conditions: str / iterable ~~ None
        Optional query portion that will be used to limit the list of results.
        If multiple conditions are provided, they are joined on an 'AND' string.
      % limit: int ~~ None
        Specifies a maximum number of items to be yielded. The limit happens on
        the database side, limiting the query results.
      % offset: int ~~ None
        Specifies the offset at which the yielded items should start. Combined
        with limit this enables proper pagination.
      % order: iterable of str/2-tuple
        Defines the fields on which the output should be ordered. This should be
        a list of strings or 2-tuples. The string or first item indicates the
        field, the second argument defines descending order (desc. if True).
      % yield_unlimited_total_first: bool ~~ False
        Instead of yielding only Record objects, the first item returned is the
        number of results from the query if it had been executed without limit.
      % search: str
        Specifies what string should be searched for in the default searchable
        database columns.
      % tables: str / iterable ~~ None
        Specifies what tables should be searched queried
      % escape: bool ~~ True
        Are conditions escaped?
      % fields: str / iterable ~~ *
        Specifies what fields should be returned
      % distinct: bool (optional).
        Performs a DISTINCT query if set to True.

    Yields:
      Record: Database record abstraction class.
    """
    if not tables:
      tables = [cls.TableName()]
    group = None
    if fields is None:
      fields = '%s.*' % cls.TableName()
    if search:
      group = '%s.%s' % (cls.TableName(), (cls.RecordKey() if getattr(cls, "RecordKey", None) else cls._PRIMARY_KEY))
      tables, searchconditions = cls._GetSearchQuery(connection, tables, search)
      if conditions:
        if type(conditions) == list:
          conditions.extend(searchconditions)
        else:
          searchconditions.append(conditions)
          conditions = searchconditions
      else:
        conditions = searchconditions
    cacheable = False
    if not offset or offset < 0:
      offset = 0
    if hasattr(cls, '_addToCache'):
      #TODO dont cache partial / multi-table objects
      cacheable = True
      connection.modelcache['_stats']['queries'].append('%s Record.List' % cls.TableName())
    with connection as cursor:
      records = cursor.Select(fields=fields,
                              table=tables, conditions=conditions,
                              limit=limit, offset=offset, order=order,
                              totalcount=yield_unlimited_total_first,
                              escape=escape, group=group, distinct=distinct)
    if yield_unlimited_total_first:
      yield records.affected
    records = [cls(connection, record) for record in list(records)]

    yield from records

    if cacheable:
      list(cls._cacheListPreseed(records))

  # SQL Records have foreign relations, saving needs an extra argument for this.
  # pylint: disable=W0221
  def Save(self, save_foreign=False):
    """Saves the changes made to the record.

    This expands on the base Save method, providing a save_foreign that will
    recursively update all nested records when set to True.

    Arguments:
      % save_foreign: bool ~~ False
        If set, each Record (subclass) contained by this one will be saved as
        well. This recursive saving triggers *before* this record itself will be
        saved. N.B. each record is saved using a separate transaction, meaning
        that a failure to save this object will *not* roll back child saves.
    """
    with self.connection as cursor:
      if save_foreign:
        self._SaveForeign(cursor)
      self._SaveSelf(cursor)
    return self
  # pylint: enable=W0221

  @classmethod
  def _GetSearchQuery(cls, connection, tables, search):
    """Extracts table information from the searchable columns list."""
    conditions = []
    like = 'like "%%%s%%"' % connection.EscapeValues(search.strip())[1:-1]
    searchconditions = []
    thistable = cls.TableName()
    for column in cls.SEARCHABLE_COLUMNS:
      if '.' not in column:
        searchconditions.append('`%s`.`%s` %s' % (thistable, column, like))
        continue
      classname, column = column.split('.', 1)
      othertable = cls._SUBTYPES[classname].TableName()
      if (othertable != thistable and
          othertable not in tables):
        fkey = cls._FOREIGN_RELATIONS.get(classname, False)
        key = othertable._PRIMARY_KEY
        if fkey and fkey.get('LookupKey', False):
          key = fkey.get('LookupKey')
        elif getattr(table, "RecordKey", None):
          key = othertable.RecordKey()
        # add the cross table join
        #TODO use referenced field, instead of just the othertable name
        conditions.append('`%s`.`%s` = `%s`.`%s' % (thistable,
            othertable,
            othertable,
            key))
        tables.append(othertable)
      searchconditions.append('`%s`.`%s` %s' % (othertable,
          column, like))

    conditions.append('(%s)' % ' or '.join(searchconditions))
    return tables, conditions

  def __json__(self, complete=True, recursive=True):
    """Returns a dictionary representation of the Record.

    Arguments:
      % complete: bool ~~ False
        Whether the foreign references on the object should all be resolved before
        converting the Record to a dictionary. Either way, existing resolved
        references will be represented as complete dictionaries.

      Returns:
        dict: dictionary representation of the record.
      """
    record_dict = {}
    record = self if complete else dict(record)
    for key, value in record.items():
      if isinstance(value, BaseRecord):
        if complete and recursive:
          record_dict[key] = value.__json__(complete=True, recursive=True)
        else:
          record_dict[key] = dict(value)
      else:
        record_dict[key] = value
    return record_dict


class VersionedRecord(Record):
  """Basic class for database table/record abstraction."""
  _LOAD_METHOD = 'FromIdentifier'
  _RECORD_KEY = None

  # ############################################################################
  # RecordKey method, analogous to TableName
  #
  @classmethod
  def RecordKey(cls):
    """Returns the record identifier fieldname.

    If this is not explicitly defined by the class constant `_RECORD_KEY`, the
    return value will be the class' TableName() with 'ID' appended.
    """
    if cls._RECORD_KEY is not None:
      return cls._RECORD_KEY
    return cls.TableName() + 'ID'

  # ############################################################################
  # Public methods for creation, deletion and storing Record objects.
  #
  @classmethod
  def FromIdentifier(cls, connection, identifier):
    """Returns the newest Record object that matches the given identifier.

    N.B. Newest is defined as 'last in lexicographical sort'.

    Arguments:
      @ connection: sqltalk.connection
        Database connection to use.
      @ identifier: obj
        The value of the record key field

    Raises:
      NotExistError:
        There is no Record that matches the given identifier.

    Returns:
      Record: The newest record for the given identifier.
    """
    safe_id = connection.EscapeValues(identifier)

    with connection as cursor:
      record = cursor.Select(
          table=cls.TableName(), order=[(cls._PRIMARY_KEY, True)],
          conditions='`%s`=%s' % (cls.RecordKey(), safe_id), limit=1)
    if not record:
      raise NotExistError('There is no %r for identifier %r' % (
          cls.__name__, identifier))
    return cls(connection, record[0])

  @classmethod
  def List(cls, connection, conditions=None, limit=None, offset=None,
           order=None, yield_unlimited_total_first=False, search=None,
           tables=None, escape=True, fields=None):
    """Yields the latest Record for each versioned entry in the table.

    Arguments:
      @ connection: object
        Database connection to use.
      % conditions: str / iterable ~~ None
        Optional query portion that will be used to limit the list of results.
        If multiple conditions are provided, they are joined on an 'AND' string.
      % limit: int ~~ None
        Specifies a maximum number of items to be yielded. The limit happens on
        the database side, limiting the query results.
      % offset: int ~~ None
        Specifies the offset at which the yielded items should start. Combined
        with limit this enables proper pagination.
      % order: iterable of str/2-tuple
        Defines the fields on which the output should be ordered. This should
        be a list of strings or 2-tuples. The string or first item indicates the
        field, the second argument defines descending order (desc. if True).
      % yield_unlimited_total_first: bool ~~ False
        Instead of yielding only Record objects, the first item returned is the
        number of results from the query if it had been executed without limit.
      % search: str
        Specifies what string should be searched for in the default searchable
        database columns.
      % tables: str / iterable ~~ None
        Specifies what tables should be searched queried
      % escape: bool ~~ True
        Are conditions escaped?
      % fields: str / iterable ~~ *
        Specifies what fields should be returned

    Yields:
      Record: The Record with the newest version for each versioned entry.
    """
    if not tables:
      tables = [cls.TableName()]
    if fields:
      if fields != '*':
        if isinstance(fields, str):
          fields = connection.EscapeField(fields)
        else:
          fields = ', '.join(connection.EscapeField(fields))
    else:
      fields = "%s.*" % cls.TableName()
    if search:
      search = search.strip()
      tables, searchconditions = cls._GetSearchQuery(connection, tables, search)
      if conditions:
        if type(conditions) == list:
          conditions.extend(searchconditions)
        else:
          searchconditions.append(conditions)
          conditions = searchconditions
      else:
        conditions = searchconditions
    field_escape = connection.EscapeField if escape else lambda x: x
    if yield_unlimited_total_first and limit is not None:
      totalcount = 'SQL_CALC_FOUND_ROWS'
    else:
      totalcount = ''
    cacheable = False
    if hasattr(cls, '_addToCache'):
      #TODO dont cache partial / multi-table objects
      cacheable = True
      connection.modelcache['_stats']['queries'].append('%s VersionedRecord.List' % cls.TableName())
    with connection as cursor:
      records = cursor.Execute("""
          SELECT %(totalcount)s %(fields)s
          FROM %(tables)s
          JOIN (SELECT MAX(`%(primary)s`) AS `max`
                FROM `%(table)s`
                GROUP BY `%(record_key)s`) AS `versions`
              ON (`%(table)s`.`%(primary)s` = `versions`.`max`)
          WHERE %(conditions)s
          %(order)s
          %(limit)s
          """ % {'totalcount': totalcount,
                 'primary': cls._PRIMARY_KEY,
                 'record_key': cls.RecordKey(),
                 'fields': fields,
                 'table': cls.TableName(),
                 'tables': cursor._StringTable(tables, field_escape),
                 'conditions': cursor._StringConditions(conditions,
                                                      field_escape),
                 'order': cursor._StringOrder(order, field_escape),
                 'limit': cursor._StringLimit(limit, offset)})
    if yield_unlimited_total_first and limit is not None:
      with connection as cursor:
        records.affected = cursor._Execute('SELECT FOUND_ROWS()')[0][0]
      yield records.affected
    # turn sqltalk rows into model
    records = [cls(connection, record) for record in list(records)]
    yield from records
    if cacheable:
      list(cls._cacheListPreseed(records))

  @classmethod
  def Versions(cls, connection, identifier, conditions='1'):
    """Yields all versions for a given record identifier.

    Arguments:
      @ connection: sqltalk.connection
        Database connection to use.
      % conditions: str
        Optional query portion that will be used to limit the list of results

    Yields:
      Record: One for each stored version for the identifier.
    """
    if isinstance(conditions, (list, tuple)):
      conditions = ' AND '.join(conditions)
    safe_id = connection.EscapeValues(identifier)
    with connection as cursor:
      records = cursor.Select(table=cls.TableName(),
                              conditions='`%s` = %s AND %s' % (
                                  cls.RecordKey(), safe_id, conditions))
    for record in records:
      yield cls(connection, record)

  # ############################################################################
  # Private methods to control VersionedRecord behaviour
  #
  @classmethod
  def _NextRecordKey(cls, cursor):
    """Returns the next record key to use, the previous (or zero) plus one."""
    return (cls._MaxRecordKey(cursor) or 0) + 1

  @classmethod
  def _MaxRecordKey(cls, cursor):
    """Returns the currently largest record key value."""
    last_key = cursor.Select(table=cls.TableName(), fields=cls.RecordKey(),
                             order=[(cls.RecordKey(), True)], limit=1)
    if last_key:
      return last_key[0][cls.RecordKey()]

  def _PreCreate(self, cursor):
    """Attaches a RecordKey to the Record if it doens't have one already.

    Before we create a new record, we need to acquire the next-in-line RecordKey
    if none has been provided. If one has been provided, we'll use that one.
    """
    if self.identifier is None:
      self.identifier = self._NextRecordKey(cursor)

  def _PreSave(self, cursor):
    """Before saving a record, reset the primary key value.

    This assures that we will not have a primary key conflict, though it does
    assume an AutoIncrement primary key field.
    """
    super(VersionedRecord, self)._PreSave(cursor)
    difference = self._Changes()
    if difference:
      self.key = None

  def _RecordUpdate(self, cursor):
    """All updates are handled as new inserts for the same Record Key."""
    difference = self._Changes()
    if difference:
      self._RecordCreate(cursor)

  # Pylint falsely believes this property is overwritten by its setter later on.
  # pylint: disable=E0202
  @property
  def identifier(self):
    """Returns the value of the version field of the record.

    This is used for the Save/Update methods, where foreign relations should be
    stored by their primary key.
    """
    return self.get(self.RecordKey())
  # pylint: enable=E0202

  # Pylint doesn't understand property setters at all.
  # pylint: disable=E0102, E0202, E1101
  @identifier.setter
  def identifier(self, value):
    """Sets the value of the primary key."""
    self[self.RecordKey()] = value
  # pylint: enable=E0102, E0202, E1101


class MongoRecord(BaseRecord):
  """Abstraction of MongoDB collection records."""
  _PRIMARY_KEY = '_id'
  _CONNECTOR = 'mongo'

  @classmethod
  def Collection(cls, connection):
    """Returns the collection that the MongoRecord resides in."""
    return getattr(connection, cls.TableName())

  @classmethod
  def Create(cls, connection, record):
    record = cls(connection, record, run_init_hook=False)
    # Accessing protected members of a foreign class.
    # pylint: disable=W0212
    record._PreCreate(None)
    record._StoreRecord()
    record._PostCreate(None)
    # pylint: enable=W0212
    return record

  @classmethod
  def DeletePrimary(cls, connection, pkey_value):
    collection = cls.Collection(connection)
    collection.remove({cls._PRIMARY_KEY: pkey_value})

  @classmethod
  def FromPrimary(cls, connection, pkey_value):
    from bson.objectid import ObjectId

    if not isinstance(pkey_value, ObjectId):
      pkey_value = ObjectId(pkey_value)
    collection = cls.Collection(connection)
    record = collection.find({cls._PRIMARY_KEY: pkey_value})
    if not record:
      raise NotExistError('There is no %r for primary key %r' % (
          cls.__name__, pkey_value))
    return cls(connection, record[0])

  @classmethod
  def List(cls, connection, conditions=None):
    for record in cls.Collection(connection).find(conditions or {}):
      yield cls(connection, record)

  def Save(self):
    changes = self._Changes()
    if changes:
      self._PreSave(None)
      self._StoreRecord()
      self._PostSave(None)
      self._record.update(changes)
    return self

  def _StoreRecord(self):
    self.key = self.Collection(self.connection).save(self._DataRecord())


def RecordTableNames():
  """Yields Record subclasses that have been defined outside this module.

  This is necessary to accurately perform automatic loading of foreign elements.
  There is one requirement to this, and that's that all subclasses of Record
  are loaded in memory by the time the first Record is instantiated, because
  this function will only be called once by default.
  """
  def GetSubTypes(cls, seen=None):
    """Recursively and depth-first retrieve subclasses of a given type."""
    seen = seen or set()
    # Pylint mistakenly believes there is no method __subclasses__
    # pylint: disable=E1101
    for sub in cls.__subclasses__():
    # pylint: enable=E1101
      if sub not in seen:
        seen.add(sub)
        yield sub
        yield from GetSubTypes(sub, seen)

  for cls in GetSubTypes(BaseRecord):
    # Do not yield subclasses defined in this module
    if cls.__module__ != __name__:
      yield cls.TableName(), cls


import functools

class CachedPage(object):
  """Abstraction class for the cached Pages table in the database."""

  MAXAGE = 61

  @classmethod
  def Clean(cls, connection, maxage=None):
    """Deletes all cached pages that are older than MAXAGE.

    An optional 'maxage' integer can be specified instead of MAXAGE.
    """
    with connection as cursor:
      cursor.Execute("""delete
        from
          %s
        where
          TIME_TO_SEC(TIMEDIFF(UTC_TIMESTAMP(), created)) > %d
          """ % (
          cls.TableName(),
          (cls.MAXAGE if maxage is None else maxage)
        ))

  @classmethod
  def FromSignature(cls, connection, maxage, name, modulename, args, kwargs):
    """Returns a cached page from the given signature."""
    with connection as cursor:
      cache = cursor.Execute("""select
          data,
          TIME_TO_SEC(TIMEDIFF(UTC_TIMESTAMP(), created)) as age,
          creating
        from
          %s
        where
          TIME_TO_SEC(TIMEDIFF(UTC_TIMESTAMP(), created)) < %d AND
          name = %s AND
          modulename = %s AND
          args = %s AND
          kwargs = %s
          order by created desc
          limit 1
          """ % (
        cls.TableName(),
        (cls.MAXAGE if maxage is None else maxage),
        connection.EscapeValues(name),
        connection.EscapeValues(modulename),
        connection.EscapeValues(args),
        connection.EscapeValues(kwargs)))

    if cache:
      if cache[0]['creating'] is not None:
        raise CurrentlyWorking(cache[0]['age'])
      return cls(connection, cache[0])
    else:
      raise cls.NotExistError('No cached data found')



"""This file holds all the decorators we use in this project."""

import codecs
from datetime import datetime
import hashlib
import pickle
import pytz
import json
import time

import uweb3
from uweb3 import model
from uweb3.request import IndexedFieldStorage

def loggedin(f):
  """Decorator that checks if the user requesting the page is logged in based on set cookie."""
  def wrapper(*args, **kwargs):
    if not args[0].user:
      return args[0].RequestLogin()
    return f(*args, **kwargs)
  return wrapper

def checkxsrf(f):
  """Decorator that checks the user's XSRF.
  The function will compare the XSRF in the user's cookie and in the
  (post) request. Make sure to have xsrf_enabled = True in the config.ini
  """
  def _clear_form_data(pagemaker):
    method = pagemaker.req.method.lower()
    # Set an attribute in the pagemaker that holds the form data on an invalid XSRF validation
    pagemaker.invalid_xsrf_data = getattr(pagemaker, method)
    # Remove the form data from the PageMaker
    setattr(pagemaker, method, IndexedFieldStorage())
    # Remove the form data from the Request class
    pagemaker.req.vars[method] = IndexedFieldStorage()
    return pagemaker

  def wrapper(*args, **kwargs):
    if args[0].req.method != "GET":
      if args[0].invalid_xsrf_token:
        _clear_form_data(args[0])
        return args[0].XSRFInvalidToken()
    return f(*args, **kwargs)
  return wrapper

def Cached(maxage=None, verbose=False, handler=None, *t_args, **t_kwargs):
  """Decorator that wraps checks the cache table for a cached page.
  The function will see if we have a recent cached output for this call,
  or if one is being created as we speak.
  Use by adding the decorator module and flagging a pagemaker function with
  it.
  from pages import decorators
  @decorators.Cached(60)
  def mypage()

  Arguments:
    #TODO: Make handler an argument instead of a kwd since it is required?
    @ handler: class CustomClass(model.Record, model.CachedPage)
      This is some sort of custom mixin class that we use to store our cached page in the database
    % maxage: int(60)
      Cache time in seconds.
    % verbose: bool(False)
      Insert html comment with cache information.
  Raises:
    KeyError
  """
  def cache_decorator(f):
    def wrapper(*args, **kwargs):
      if not handler:
        raise KeyError("A handler is required for storing this page into the database.")
      create = False
      name = f.__name__
      modulename = f.__module__
      handler.Clean(args[0].connection, maxage)
      requesttime = time.time()
      time.clock()
      sleep = 0.3
      maxsleepinterval = 2
      try:  # see if we have a cached version thats not too old
        data = handler.FromSignature(args[0].connection,
                                              maxage,
                                              name, modulename,
                                              json.dumps(args[1:]),
                                              json.dumps(kwargs))
        if verbose:
          data = '%s<!-- cached %ds ago -->' % (
                                             pickle.loads(codecs.decode(data['data'].encode(), "base64")),
                                             data['age'])
        else:
          data = pickle.loads(codecs.decode(data['data'].encode(), "base64"))
      except model.CurrentlyWorking:  # we dont have anything fresh enough, but someones working on it
        age = 0
        while age < maxage:  # as long as there's no output, we should try periodically until we have waited too long
          time.sleep(sleep)
          age = (time.time() - requesttime)
          try:
            data = handler.FromSignature(args[0].connection,
                                                  maxage,
                                                  name, modulename,
                                                  json.dumps(args[1:]),
                                                  json.dumps(kwargs))
            break
          except Exception:
            sleep = min(sleep*2, maxsleepinterval)
        try:
          data = pickle.loads(codecs.decode(data['data'].encode(), "base64"))
          if verbose:
            data += '<!-- waited for fresh content for %s seconds -->' % age
        except NameError:
          create = True
      except uweb3.model.NotExistError:  # we don't have anything fresh enough, lets create
        create = True
      if create:
        try:
          now = str(pytz.utc.localize(datetime.utcnow()))[0:19]
          # create the db row for this call, let other processes know we are working on it.
          cache = handler.Create(args[0].connection, {
            'name': name,
            'modulename': modulename,
            'args': json.dumps(args[1:]),
            'kwargs': json.dumps(kwargs),
            'creating': now,
            'created': now
            })
          data = f(*args, **kwargs)
          cache['data'] = codecs.encode(pickle.dumps(data), "base64").decode()
          # update the created time to now, as we are done.
          cache['created'] = str(pytz.utc.localize(datetime.utcnow()))[0:19]
          cache['creating'] = None
          cache.Save()
          if verbose:
            data += '<!-- Freshly generated -->'
        except Exception: #This is probably a pymysql Error. or db collision, whilst unfortunate, we wont break the page on this
          pass
      return data
    return wrapper
  return cache_decorator

def ContentType(content_type):
  """Decorator that wraps and returns sets the contentType."""
  def content_type_decorator(f):
    def wrapper(*args, **kwargs):
      pageresult = f(*args, **kwargs) or {}
      if not isinstance(pageresult, uweb3.Response):
        return uweb3.Response(pageresult,
                              content_type=content_type)
      if isinstance(pageresult, uweb3.Response):
        pageresult.content_type = content_type
      args[0].req.content_type = content_type
      return pageresult
    return wrapper
  return content_type_decorator

def CSP(resourcetype, urls, append=True):
  """Decorator that injects a new CSP allowed source into the current csp output."""
  def csp_decorator(f):
    def wrapper(*args, **kwargs):
      args[0]._SetCsp(resourcetype, urls, append)
      return f(*args, **kwargs) or {}
    return wrapper
  return csp_decorator

def TemplateParser(template, *t_args, **t_kwargs):
  """Decorator that wraps and returns the output.

  The output is wrapped in a templateparser call if its not already something
  that we prepared for direct output to the client.
  """
  def template_decorator(f):
    def wrapper(*args, **kwargs):
      pageresult = f(*args, **kwargs) or {}
      if not isinstance(pageresult, (str, uweb3.Response, uweb3.Redirect)):
        return args[0].parser.Parse(template, **pageresult)
      return pageresult
    return wrapper
  return template_decorator



#!/usr/bin/python3
"""This file contains the connector for Signed cookies."""
__author__ = 'Jan Klopper (janunderdark.nl)'
__version__ = 0.1

import os
from base64 import b64encode

from . import Connector

class SignedCookie(Connector):
  """Adds a signed cookie connection to the connection manager object.

  The name of the class is used as the Cookiename"""

  PERSISTENT = False

  def __init__(self, config, options, request, debug=False):
    """Sets up the local connection to the signed cookie store, and generates a
    new secret key if no key can be found in the config"""
    self.debug = debug
    # Generating random seeds on uWeb3 startup or fetch from config
    try:
      self.options = options[self.Name()]
      self.secure_cookie_secret = self.options['secret']
    except KeyError:
      secret = self.GenerateNewKey()
      config.Create(self.Name(), 'secret', secret)
      if self.debug:
        print('SignedCookie: Wrote new secret random to config.')
      self.secure_cookie_secret = secret
    self.connection = (request, request.vars['cookie'], self.secure_cookie_secret)

  @staticmethod
  def GenerateNewKey(length=128):
    return b64encode(os.urandom(length)).decode('utf-8')



#!/usr/bin/python3
# -*- coding: utf-8 -*-
"""µWeb3 Framework"""

__version__ = '3.0.4'

# Standard modules
import configparser
import logging
import os
import re
import sys
from wsgiref.simple_server import make_server
import datetime

# Package modules
from . import pagemaker, request

# Package classes
from .response import Response, Redirect
from .pagemaker import PageMaker, decorators, WebsocketPageMaker, DebuggingPageMaker, LoginMixin
from .model import SettingsManager
from .libs.safestring import HTMLsafestring, JSONsafestring, JsonEncoder, Basesafestring
from importlib import reload

class Error(Exception):
  """Superclass used for inheritance and external exception handling."""

class ImmediateResponse(Exception):
  """Used to trigger an immediate response, foregoing the regular returns."""

class HTTPException(Error):
  """SuperClass for HTTP exceptions."""

class HTTPRequestException(HTTPException):
  """Exception for http request errors."""

class NoRouteError(Error):
  """The server does not know how to route this request"""

class Registry:
  """Something to hook stuff to"""


class Router:
  def __init__(self, page_class):
    self.pagemakers = page_class.LoadModules()
    self.pagemakers.append(page_class)

  def router(self, routes):
    """Returns the first request handler that matches the request URL.

    The `routes` argument is an iterable of 2-tuples, each of which contain a
    pattern (regex) and the name of the handler to use for matching requests.

    Before returning the closure, all regexp are compiled, and handler methods
    are retrieved from the provided `page_class`.

    Arguments:
      @ routes: iterable of 2-tuples.
        Each tuple is a pair of `pattern` and `handler`, both are strings.

    Returns:
      request_router: Configured closure that processes urls.
    """
    req_routes = []
    # Variable used to store websocket pagemakers,
    # these pagemakers are only created at startup but can have multiple routes.
    # To prevent creating the same instance for each route we store them in a dict
    websocket_pagemaker = {}
    for pattern, *details in routes:
      page_maker = None
      for pm in self.pagemakers:
        # Check if the page_maker has the method/handler we are looking for
        if hasattr(pm, details[0]):
          page_maker = pm
          break
      if callable(pattern):
        # Check if the page_maker is already in the dict, if not instantiate
        # if so just use that one. This prevents creating multiple instances for one route.
        if not websocket_pagemaker.get(page_maker.__name__):
          websocket_pagemaker[page_maker.__name__] = page_maker()
        pattern(getattr(websocket_pagemaker[page_maker.__name__], details[0]))
        continue
      if not page_maker:
        raise NoRouteError(f"µWeb3 could not find a route handler called '{details[0]}' in any of the PageMakers, your application will not start.")
      req_routes.append((re.compile(pattern + '$', re.UNICODE),
                        details[0], #handler,
                        details[1].upper() if len(details) > 1 else 'ALL', #request types
                        details[2].lower() if len(details) > 2 else '*', #host
                        page_maker #pagemaker class
                        ))

    def request_router(url, method, host):
      """Returns the appropriate handler and arguments for the given `url`.

      The`url` is matched against the compiled patterns in the `req_routes`
      provided by the outer scope. Upon finding a pattern that matches, the
      match groups from the regex and the unbound handler method are returned.

      N.B. The rules are such that the first matching route will be used. There
      is no further concept of specificity. Routes should be written with this in
      mind.

      Arguments:
        @ url: str
          The URL requested by the client.
        @ method: str
          The http method requested by the client.
        @ host: str
          The http host header value requested by the client.

      Raises:
        NoRouteError: None of the patterns match the requested `url`.

      Returns:
        2-tuple: handler method (unbound), and tuple of pattern matches.
      """

      for pattern, handler, routemethod, hostpattern, page_maker in req_routes:
        if routemethod != 'ALL':
          # clearly not the route we where looking for
          if isinstance(routemethod, tuple) and method not in routemethod:
            continue
          if method != routemethod:
            continue

        hostmatch = None
        if hostpattern != '*':
          # see if we can match this host and extact any info from it.
          hostmatch = re.compile(f"^{host}$").match(hostpattern)
          if not hostmatch:
            # clearly not the host we where looking for
            continue
          hostmatch = hostmatch.groups()
        match = pattern.match(url)
        if match:
          # strip out optional groups, as they return '', which would override
          # the handlers default argument values later on in the page_maker
          groups = (group for group in match.groups() if group)
          return handler, groups, hostmatch, page_maker
      raise NoRouteError(url +' cannot be handled')
    return request_router


class uWeb:
  """Returns a configured closure for handling page requests.

  This closure is configured with a precomputed set of routes and handlers using
  the Router function. After this, incoming requests are processed and delegated
  to the correct PageMaker handler.

  The url in the received `req` object is taken and matches against the
  `router`` (refer to Router() for more documentation on this).


  Takes:
    @ page_class: PageMaker
      Class that holds request handling methods as defined in the `routes`
    @ router: request router
      The result of the Router() function.
    @ config: dict
      Configuration for the PageMaker. Typically contains entries for database
      connections, default search paths etc.

  Returns:
    RequestHandler: Configured closure that is ready to process requests.
  """
  def __init__(self, page_class, routes, executing_path=None, config='config'):
    self.executing_path = executing_path or os.path.dirname(__file__)
    self.config = SettingsManager(filename=config, path=self.executing_path)
    self.logger = self.setup_logger()
    self.inital_pagemaker = page_class
    self.registry = Registry()
    self.registry.logger = logging.getLogger('root')
    self.router = Router(page_class).router(routes)
    self.setup_routing()
    self.encoders = {
        'text/html': lambda x: HTMLsafestring(x, unsafe=True),
        'text/plain': str,
        'application/json': lambda x: JSONsafestring(x, unsafe=True),
        'default': str,}

  def __call__(self, env, start_response):
    """WSGI request handler.
    Accepts the WSGI `environment` dictionary and a function to start the
    response and returns a response iterator.
    """
    req = request.Request(env, self.registry)
    req.env['REAL_REMOTE_ADDR'] = request.return_real_remote_addr(req.env)
    response = None
    method = '_NotFound'
    args = None
    rollback = False
    try:
      method, args, hostargs, page_maker = self.router(req.path,
                                            req.env['REQUEST_METHOD'],
                                            req.env['host'])
    except NoRouteError:
      # When we catch this error this means there is no method for the route in the currently selected pagemaker.
      # If this happens we default to the initial pagemaker because we don't know what the target pagemaker should be.
      # Then we set an internalservererror and move on
      page_maker = self.inital_pagemaker
    try:
      # instantiate the pagemaker for this request
      pagemaker_instance = page_maker(req,
                            config=self.config,
                            executing_path=self.executing_path)
      # specifically call _PreRequest as promised in documentation
      if hasattr(pagemaker_instance, '_PreRequest'):
        pagemaker_instance = pagemaker_instance._PreRequest()

      response = self.get_response(pagemaker_instance, method, args)
    except Exception:
      # something broke in our pagemaker_instance, lets fall back to the most basic pagemaker for error output
      if hasattr(pagemaker_instance, '_ConnectionRollback'):
        try:
          pagemaker_instance._ConnectionRollback()
        except:
          pass
      pagemaker_instance = PageMaker(req,
                            config=self.config,
                            executing_path=self.executing_path)
      response = pagemaker_instance.InternalServerError(*sys.exc_info())

    static = False
    if method == 'Static':
      static = True

    if not static:
      if not isinstance(response, Response):
        # print('Upgrade response to Response class: %s' % type(response))
        req.response.text = response
        response = req.response

      if not isinstance(response.text, Basesafestring):
        # make sure we always output Safe Strings for our known content-types
        encoder = self.encoders.get(response.clean_content_type(), self.encoders['default'])
        response.text = encoder(response.text)

      if hasattr(pagemaker_instance, '_PostRequest'):
        pagemaker_instance._PostRequest()

    # CSP might be unneeded for some static content,
    # https://github.com/w3c/webappsec/issues/520
    if hasattr(pagemaker_instance, '_CSPheaders'):
      pagemaker_instance._CSPheaders()

    # provide users with a _PostRequest method to overide too
    if not static and hasattr(pagemaker_instance, 'PostRequest'):
      response = pagemaker_instance.PostRequest(response)

    # we should at least send out something to make sure we are wsgi compliant.
    if not response.text:
      response.text = ''

    self._logging(req, response)
    start_response(response.status, response.headerlist)
    try:
      yield response.text.encode(response.charset)
    except AttributeError:
      yield response.text

  def setup_logger(self):
    logger = logging.getLogger('uweb3_logger')
    logger.setLevel(logging.INFO)
    fh = logging.FileHandler(os.path.join(self.executing_path, 'access_logging.log'))
    fh.setLevel(logging.INFO)
    logger.addHandler(fh)
    return logger

  def _logging(self, req, response):
    """Logs incoming requests to a logfile.
    This is enabled by default, even if its missing in the config file.
    """
    if (self.config.options.get('development', None) and
        self.config.options['development'].get('access_logging', True) == 'False'):
      return

    host = req.env['HTTP_HOST'].split(':')[0]
    date = datetime.datetime.now().strftime('%d/%m/%Y %H:%M:%S')
    method = req.method
    path = req.path
    status = response.httpcode
    protocol = req.env.get('SERVER_PROTOCOL')
    self.logger.info(f"""{host} - - [{date}] \"{method} {path} {status} {protocol}\"""")

  def get_response(self, page_maker, method, args):
    try:
      if method != 'Static':
        # We're specifically calling _PostInit here as promised in documentation.
        # pylint: disable=W0212
        page_maker._PostInit()
      elif hasattr(page_maker, '_StaticPostInit'):
        # We're specifically calling _StaticPostInit here as promised in documentation, seperate from the regular PostInit to keep things fast for static pages
        page_maker._StaticPostInit()

      # pylint: enable=W0212
      return getattr(page_maker, method)(*args)
    except pagemaker.ReloadModules as message:
      reload_message = reload(sys.modules[self.inital_pagemaker.__module__])
      return Response(content='%s\n%s' % (message, reload_message))
    except ImmediateResponse as err:
      return err[0]
    except Exception:
      if (self.config.options.get('development', False) and
          self.config.options['development'].get('error_logging', False) == 'True'):
        logger = logging.getLogger('uweb3_exception_logger')
        fh = logging.FileHandler(os.path.join(self.executing_path, 'uweb3_uncaught_exceptions.log'))
        logger.addHandler(fh)
        logger.exception("UNCAUGHT EXCEPTION:")
      return page_maker.InternalServerError(*sys.exc_info())

  def serve(self):
    """Sets up and starts WSGI development server for the current app."""
    host = 'localhost'
    port = 8001
    hotreload = False
    interval = None
    ignored_directories = ['__pycache__',
                           self.inital_pagemaker.PUBLIC_DIR,
                           self.inital_pagemaker.TEMPLATE_DIR]
    ignored_extensions = []
    if self.config.options.get('development', False):
      host = self.config.options['development'].get('host', host)
      port = self.config.options['development'].get('port', port)
      hotreload = self.config.options['development'].get('reload', False) in ('True', 'true')
      interval = int(self.config.options['development'].get('checkinterval', 0))
      if 'ignored_extensions' in self.config.options['development']:
        ignored_extensions = self.config.options['development'].get('ignored_extensions', '').split(',')
      if 'ignored_directories' in self.config.options['development']:
        ignored_directories += self.config.options['development'].get('ignored_directories', '').split(',')
    server = make_server(host, int(port), self)
    print(f'Running µWeb3 server on http://{server.server_address[0]}:{server.server_address[1]}')
    print(f'Root dir is: {self.executing_path}')
    try:
      if hotreload:
        print(f'Hot reload is enabled for changes in: {self.executing_path}')
        HotReload(self.executing_path, interval=interval,
            ignored_extensions=ignored_extensions,
            ignored_directories=ignored_directories)
      server.serve_forever()
    except Exception as error:
      print(error)
      server.shutdown()

  def setup_routing(self):
    if isinstance(self.inital_pagemaker, list):
      routes = [route for route in self.inital_pagemaker[1:]]
      self.inital_pagemaker[0].AddRoutes(tuple(routes))
      self.inital_pagemaker = self.inital_pagemaker[0]

    default_route = "routes"
    automatic_detection = True
    if self.config.options.get('routing'):
      default_route = self.config.options['routing'].get('default_routing', default_route)
      automatic_detection = self.config.options['routing'].get('disable_automatic_route_detection', 'False') != 'True'

    if automatic_detection:
      self.inital_pagemaker.LoadModules(routes=default_route)


class HotReload:
    """This class handles the thread which scans for file changes in the
    execution path and restarts the server if needed"""
    IGNOREDEXTENSIONS = [".pyc", '.ini', '.md', '.html', '.log', '.sql']

    def __init__(self, path, interval=None, ignored_extensions=None, ignored_directories=None):
      """Takes a path, an optional interval in seconds and an optional flag
      signaling a development environment which will set the path for new and
      changed file checking on the parent folder of the serving file."""
      import threading
      self.running = threading.Event()
      self.interval = interval or 1
      self.path = os.path.dirname(path)
      self.ignoredextensions = self.IGNOREDEXTENSIONS + (ignored_extensions or [])
      self.ignoreddirectories = ignored_directories
      self.thread = threading.Thread(target=self.Run)
      self.thread.daemon = True
      self.thread.start()

    def Run(self):
      """ Method runs forever and watches all files in the project folder."""
      self.watched_files = self.Files()
      self.mtimes = [(f, os.path.getmtime(f)) for f in self.watched_files]

      import time
      while True:
        time.sleep(self.interval)
        new = self.Files(self.watched_files)
        if new:
          print('{color}New file added or deleted\x1b[0m \nRestarting µWeb3'.format(color='\x1b[7;30;41m'))
          self.Restart()
        for f, mtime in self.mtimes:
          if os.path.getmtime(f) != mtime:
            print('{color}Detected changes in {file}\x1b[0m \nRestarting µWeb3'.format(color='\x1b[7;30;41m', file=f))
            self.Restart()

    def Files(self, current=None):
      """Returns all files inside the working directory of uweb3."""
      if not current:
        current = set()
      new = set()
      for dirpath, dirnames, filenames in os.walk(self.path):
        if any(list(map(lambda dirname: dirname in dirpath, self.ignoreddirectories))):
          continue
        for file in filenames:
          fullname = os.path.join(dirpath, file)
          if fullname in current or fullname.endswith('~'):
            continue
          ext = os.path.splitext(file)[1]
          if ext not in self.ignoredextensions:
            new.add(fullname)
      return new

    def Restart(self):
      """Restart uweb3 with all provided system arguments."""
      self.running.clear()
      os.execl(sys.executable, sys.executable, * sys.argv)



#!/usr/bin/python3
"""Test suite for the database abstraction module (model)."""

# Too many public methods
# pylint: disable=R0904

# Standard modules
import unittest

# Importing uWeb3 makes the SQLTalk library available as a side-effect
from uweb3.libs.sqltalk import mysql
# Unittest target
from uweb3 import model
from pymysql.err import InternalError

# ##############################################################################
# Record classes for testing
#
class BasicTestRecord(model.Record):
  """Test record for offline tests."""


class Author(model.Record):
  """Author class for testing purposes."""


class Book(model.Record):
  """Book class for testing purposes."""


class Writer(model.Record):
  """Writer class for testing purposes, will manage `writers` table."""
  _TABLE = 'writers'


class VersionedAuthor(model.VersionedRecord):
  """Versioned author table for testing purposes."""


class VersionedBook(model.VersionedRecord):
  """Versioned Book class for testing purposes."""


class Compounded(model.Record):
  """Compound key record for generic storage."""
  _PRIMARY_KEY = 'first', 'second'


# ##############################################################################
# Start of tests
#
class BaseRecordTests(unittest.TestCase):
  """Offline tests of methods and behavior of the BaseRecord class."""
  def setUp(self):
    """Sets up the tests for the offline Record test."""
    self.record_class = BasicTestRecord

  def testTableName(self):
    """[BaseRecord] TableName returns the expected value and obeys _TABLE"""
    self.assertEqual(self.record_class.TableName(), 'basicTestRecord')
    self.record_class._TABLE = 'WonderfulSpam'
    self.assertEqual(self.record_class.TableName(), 'WonderfulSpam')

  def testPrimaryKey(self):
    """[BaseRecord] Primary key value on `key` property, default field 'ID'"""
    record = self.record_class(None, {'ID': 12, 'name': 'J.R.R. Tolkien'})
    self.assertEqual(record.key, 12)

  def testPrimaryKeyChanges(self):
    """[BaseRecord] Defining _PRIMARY_KEY overrides default value"""
    record = self.record_class(None, {'ID': 12, 'name': 'K. May'})
    self.record_class._PRIMARY_KEY = 'name'
    self.assertEqual(record.key, 'K. May')

  def testEquality(self):
    """[BaseRecord] Records of the same content are equal to eachother"""
    record_one = self.record_class(None, {'ID': 2, 'name': 'Rowling'})
    record_two = self.record_class(None, {'ID': 2, 'name': 'Rowling'})
    record_three = self.record_class(None, {'ID': 3, 'name': 'Rowling'})
    record_four = self.record_class(None, {'ID': 2, 'name': 'Rowling', 'x': 2})
    self.assertFalse(record_one is record_two)
    self.assertEqual(record_one, record_two)
    self.assertNotEqual(record_one, record_three)
    self.assertNotEqual(record_one, record_four)


class RecordTests(unittest.TestCase):
  """Online tests of methods and behavior of the Record class."""
  def setUp(self):
    """Sets up the tests for the Record class."""
    self.connection = DatabaseConnection()
    with self.connection as cursor:
      cursor.Execute("""CREATE TABLE `author` (
                            `ID` smallint(5) unsigned NOT NULL AUTO_INCREMENT,
                            `name` varchar(32) NOT NULL,
                            PRIMARY KEY (`ID`)
                          ) ENGINE=InnoDB  DEFAULT CHARSET=utf8""")
      cursor.Execute("""CREATE TABLE `book` (
                            `ID` smallint(5) unsigned NOT NULL AUTO_INCREMENT,
                            `author` smallint(5) unsigned NOT NULL,
                            `title` varchar(32) NOT NULL,
                            PRIMARY KEY (`ID`)
                          ) ENGINE=InnoDB DEFAULT CHARSET=utf8""")

  def tearDown(self):
    """Destroy tables after testing."""
    with self.connection as cursor:
      cursor.Execute('DROP TABLE `author`')
      cursor.Execute('DROP TABLE `book`')

  def testLoadPrimary(self):
    """[Record] Records can be loaded by primary key using FromPrimary()"""
    with self.connection as cursor:
      inserted = cursor.Insert(table='author', values={'name': 'A. Chrstie'})
    author = Author.FromPrimary(self.connection, inserted.insertid)
    self.assertEqual(type(author), Author)
    self.assertEqual(len(author), 2)
    self.assertEqual(author.key, inserted.insertid)
    self.assertEqual(author['name'], 'A. Chrstie')

  def testLoadPrimaryWithChangedKey(self):
    """[Record] Records can be loaded from alternative primary key"""
    with self.connection as cursor:
      inserted = cursor.Insert(table='author', values={'name': 'B. Cartland'})
    # Adjust primary key field name
    Author._PRIMARY_KEY = 'name'
    # Actual tests
    author = Author.FromPrimary(self.connection, 'B. Cartland')
    self.assertEqual(type(author), Author)
    self.assertEqual(len(author), 2)
    self.assertEqual(author.key, author['name'])
    self.assertEqual(author['ID'], inserted.insertid)
    # Restore global state
    Author._PRIMARY_KEY = 'ID'

  def testCreateRecord(self):
    """Database records can be created using Create()"""
    new_author = Author.Create(self.connection, {'name': 'W. Shakespeare'})
    author = Author.FromPrimary(self.connection, new_author.key)
    self.assertEqual(author['name'], 'W. Shakespeare')

  def testCreateRecordWithBadField(self):
    """Database record creation fails if there are unknown fields present"""
    self.assertRaises(InternalError, Author.Create, self.connection,
                      {'name': 'L. Tolstoy', 'email': 'leo@tolstoy.ru'})

  def testUpdateRecord(self):
    """The record can be given new values and these are properly stored"""
    author = Author.Create(self.connection, {'name': 'B. King'})
    author['name'] = 'S. King'
    author.Save()
    same_author = Author.FromPrimary(self.connection, 1)
    self.assertEqual(same_author['name'], 'S. King')
    self.assertEqual(same_author, author)

  def testUpdateRecordWithBadField(self):
    """Database record updating fails if there are unknown fields present"""
    author = Author.Create(self.connection, {'name': 'A. Pushkin'})
    author['specialty'] = 'poetry'
    self.assertRaises(InternalError, author.Save)

  def testUpdatePrimaryKey(self):
    """Saving with an updated primary key properly saved the record"""
    author = Author.Create(self.connection, {'name': 'C. Dickens'})
    self.assertEqual(author.key, 1)
    author['ID'] = 101
    author.Save()
    self.assertRaises(model.NotExistError, Author.FromPrimary,
                      self.connection, 1)
    same_author = Author.FromPrimary(self.connection, 101)
    self.assertEqual(same_author, author)

  def testLoadRelated(self):
    """Fieldnames that match tablenames trigger automatic loading"""
    Author.Create(self.connection, {'name': 'D. Koontz'})
    book = Book(self.connection, {'author': 1})
    self.assertEqual(type(book['author']), Author)
    self.assertEqual(book['author']['name'], 'D. Koontz')
    self.assertEqual(book['author'].key, 1)

  def testLoadRelatedFailure(self):
    """Automatic loading raises NotExistError if the foreign record is absent"""
    book = Book(self.connection, {'author': 1})
    self.assertRaises(model.NotExistError, book.__getitem__, 'author')

  def testLoadRelatedSuppressedForNone(self):
    """Automatic loading is not attempted when the field value is `None`"""
    book = Book(self.connection, {'author': None})
    self.assertEqual(book['author'], None)


class NonStandardTableAndRelations(unittest.TestCase):
  """Verified autoloading works for records with an alternate table name."""
  def setUp(self):
    """Sets up the tests for the Record class."""
    self.connection = DatabaseConnection()
    with self.connection as cursor:
      cursor.Execute("""CREATE TABLE `writers` (
                            `ID` smallint(5) unsigned NOT NULL AUTO_INCREMENT,
                            `name` varchar(32) NOT NULL,
                            PRIMARY KEY (`ID`)
                          ) ENGINE=InnoDB  DEFAULT CHARSET=utf8""")
      cursor.Execute("""CREATE TABLE `book` (
                            `ID` smallint(5) unsigned NOT NULL AUTO_INCREMENT,
                            `author` smallint(5) unsigned NOT NULL,
                            `title` varchar(32) NOT NULL,
                            PRIMARY KEY (`ID`)
                          ) ENGINE=InnoDB DEFAULT CHARSET=utf8""")

  def tearDown(self):
    """Destroy tables after testing."""
    with self.connection as cursor:
      cursor.Execute('DROP TABLE `writers`')
      cursor.Execute('DROP TABLE `book`')

  def testVerifyNoLoad(self):
    """No loading is performed on a field that matches a class but no table"""
    book = Book(self.connection, {'writer': 1, 'title': 'Trouble Shooter'})
    self.assertEquals(book['writer'], 1)

  def testVerifyFailedLoad(self):
    """Loading is attempted for the field name matching the Record's table """
    book = Book(self.connection, {'writers': 1, 'title': 'Hondo'})
    self.assertRaises(model.NotExistError, book.__getitem__, 'writers')

  def testSuccessfulLoadWithTableName(self):
    """Loading works from the adjusted table name"""
    author = Writer.Create(self.connection, {'name': 'R. Ludlum'})
    book = Book(self.connection, {'writers': 1, 'title': 'Bourne Identity'})
    self.assertEquals(book['writers'], author)

  def testLoadWithForeignRelationMapping(self):
    """Loading from alternative fieldname->table relation works"""
    author = Writer.Create(self.connection, {'name': 'R.L. Stine'})
    book = Book.Create(self.connection, {'author': 1, 'title': 'Fright Camp'})
    self.assertRaises(self.connection.ProgrammingError,
                      book.__getitem__, 'author')  # No table `author`
    Book._FOREIGN_RELATIONS = {'author': Writer}
    self.assertEqual(book['author'], author)
    del Book._FOREIGN_RELATIONS  # Don't persist changes to global state


class VersionedRecordTests(unittest.TestCase):
  """Tests for the VersionedRecord class."""
  def setUp(self):
    """Sets up the tests for the VersionedRecord class."""
    self.connection = DatabaseConnection()
    with self.connection as cursor:
      cursor.Execute("""CREATE TABLE `versionedAuthor` (
                            `ID` smallint(5) unsigned NOT NULL AUTO_INCREMENT,
                            `versionedAuthorID` smallint(5) unsigned NOT NULL,
                            `name` varchar(32) NOT NULL,
                            PRIMARY KEY (`ID`),
                            KEY `recordKey` (`versionedAuthorID`)
                          ) ENGINE=InnoDB DEFAULT CHARSET=utf8""")
      cursor.Execute("""CREATE TABLE `versionedBook` (
                            `ID` smallint(5) unsigned NOT NULL AUTO_INCREMENT,
                            `versionedBookID` smallint(5) unsigned NOT NULL,
                            `versionedAuthor` smallint(5) unsigned NOT NULL,
                            `title` varchar(32) NOT NULL,
                            PRIMARY KEY (`ID`),
                            KEY `recordKey` (`versionedBookID`)
                          ) ENGINE=InnoDB DEFAULT CHARSET=utf8""")

  def tearDown(self):
    """Destroy tables after testing."""
    with self.connection as cursor:
      cursor.Execute('DROP TABLE `versionedAuthor`')
      cursor.Execute('DROP TABLE `versionedBook`')

  def testRecordKeyName(self):
    """[Versioned] Versioning key name follows table name unless specified"""
    # Accessing protected members to check intended behavior
    # pylint: disable=W0212
    # Sanity checks, we're changing global scope here
    self.assertTrue(VersionedAuthor._TABLE is None)
    self.assertTrue(VersionedAuthor._RECORD_KEY is None)
    # Actual tests
    self.assertEqual(VersionedAuthor.RecordKey(), 'versionedAuthorID')
    VersionedAuthor._TABLE = 'author'
    self.assertEqual(VersionedAuthor.RecordKey(), 'authorID')
    VersionedAuthor._RECORD_KEY = 'recordKey'
    self.assertEqual(VersionedAuthor.RecordKey(), 'recordKey')
    # Restore global state
    VersionedAuthor._TABLE = None
    VersionedAuthor._RECORD_KEY = None

  def testCreateVersioned(self):
    """[Versioned] Creating and loading a record from identifier works"""
    author = VersionedAuthor.Create(self.connection, {'name': 'J. Grisham'})
    loaded = VersionedAuthor.FromIdentifier(self.connection, author.identifier)
    self.assertEqual(loaded['name'], 'J. Grisham')
    self.assertEqual(loaded, author)

  def testUpdateVersioned(self):
    """[Versioned] Updating records and loading from identifier works"""
    author = VersionedAuthor.Create(self.connection, {'name': 'Z. Gray'})
    initial_primary = author.key
    author['name'] = 'Z. Grey'
    author.Save()
    self.assertNotEqual(author.key, initial_primary)
    # Loading from identifier gives updated name
    loaded = VersionedAuthor.FromIdentifier(self.connection, author.identifier)
    self.assertEqual(loaded['name'], 'Z. Grey')
    # Loading from old primary key gives old name
    loaded = VersionedAuthor.FromPrimary(self.connection, initial_primary)
    self.assertEqual(loaded['name'], 'Z. Gray')

  def testListVersions(self):
    """[Versioned] Listing versions works, and happens in [old]-->[new] order"""
    author = VersionedAuthor.Create(self.connection, {'name': 'A. Martin'})
    author['name'] = 'A. Rice'
    author.Save()
    versions = list(VersionedAuthor.Versions(self.connection,
                                             author.identifier))
    self.assertEqual(len(versions), 2)
    self.assertEqual(versions[0]['name'], 'A. Martin')
    self.assertEqual(versions[1]['name'], 'A. Rice')

  def testRelationsBasedOnIdentifier(self):
    """[Versioned] Related loading defaults to using FromIdentifier"""
    # Set up records with different record keys and identifiers
    collins = VersionedAuthor.Create(self.connection, {'name': 'K. Collins'})
    collins['name'] = 'J. Collins'
    collins.Save()
    patten = VersionedAuthor.Create(self.connection, {'name': 'G. Patten'})
    # Verify sanity of keys and identifiers
    self.assertEqual(collins.key, 2)
    self.assertEqual(collins.identifier, 1)
    self.assertEqual(patten.key, 3)
    self.assertEqual(patten.identifier, 2)
    # Create book with foreign relation to author and perform actual test
    book = VersionedBook(self.connection, {
        'title': 'The Diamond Sport', 'versionedAuthor': 2})
    self.assertEqual(book['versionedAuthor'], patten)

  def testRelationsWithModifiedLoadRelationsMethod(self):
    """[Versioned] Related loading can be controlled with _LOAD_METHOD"""
    # Accessing protected members to verify and modify behavior
    # pylint: disable=W0212
    # Sanity checks, we're changing global scope here
    self.assertEqual(VersionedBook._LOAD_METHOD, 'FromIdentifier')
    # Actual tests
    VersionedAuthor._LOAD_METHOD = 'FromPrimary'
    author = VersionedAuthor.Create(self.connection, {'name': 'L. Amour'})
    author['name'] = "L. L'Amour"
    author.Save()
    book = VersionedBook(self.connection, {
        'title': 'The Riders of High Rock', 'versionedAuthor': 1})
    self.assertEqual(book['versionedAuthor']['name'], 'L. Amour')
    latest_version = VersionedAuthor.FromIdentifier(self.connection, 1)
    self.assertEqual(latest_version['name'], "L. L'Amour")
    # Restore global state
    VersionedBook._LOAD_METHOD = 'FromIdentifier'

  def testRelationsUsingCustomForeignRelations(self):
    """[Versioned] Related loading method can be set with _FOREIGN_RELATIONS"""
    # Accessing protected members to verify and modify behavior
    # pylint: disable=W0212
    # Sanity checks, we're changing global scope here
    self.assertEqual(VersionedBook._FOREIGN_RELATIONS, {})
    # Actual tests
    VersionedBook._FOREIGN_RELATIONS = {'versionedAuthor': {
        'class': 'VersionedAuthor', 'loader': 'FromPrimary'}}
    author = VersionedAuthor.Create(self.connection, {'name': 'H. Alger'})
    author['name'] = 'H. Alger, Jr.'
    author.Save()
    book = VersionedBook(self.connection, {
        'title': 'Voices of the Past', 'versionedAuthor': 1})
    self.assertEqual(book['versionedAuthor']['name'], 'H. Alger')
    latest_version = VersionedAuthor.FromIdentifier(self.connection, 1)
    self.assertEqual(latest_version['name'], 'H. Alger, Jr.')
    # Restore global state
    VersionedBook._FOREIGN_RELATIONS = {}


class CompoundKeyRecordTests(unittest.TestCase):
  """Tests for Record classes with a compound key."""
  def setUp(self):
    """Sets up the tests for the VersionedRecord class."""
    self.connection = DatabaseConnection()
    with self.connection as cursor:
      cursor.Execute("""CREATE TABLE `compounded` (
                            `first` smallint(5) unsigned NOT NULL AUTO_INCREMENT,
                            `second` smallint(5) unsigned NOT NULL,
                            `message` varchar(32) NOT NULL,
                            PRIMARY KEY (`first`, `second`)
                          ) ENGINE=InnoDB DEFAULT CHARSET=utf8""")

  def tearDown(self):
    """Destroy tables after testing."""
    with self.connection as cursor:
      cursor.Execute('DROP TABLE `compounded`')

  def testCreate(self):
    """[Compound] Creating a compound record requires both keys provided"""
    compound = Compounded.Create(self.connection, {
        'first': 1, 'second': 1, 'message': 'New compound key record'})
    self.assertEqual(compound.key, (1, 1))

  def testLoadPrimary(self):
    Compounded.Create(self.connection, {
        'first': 12, 'second': 42, 'message': 'Ahoi Ahoi'})
    compound = Compounded.FromPrimary(self.connection, (12, 42))
    self.assertEqual(compound['message'], 'Ahoi Ahoi')

  def testLoadWrongValueCount(self):
    """[Compound] Loading from primary requires the correct number of values"""
    self.assertRaises(
        TypeError, Compounded.FromPrimary, self.connection, 12)
    self.assertRaises(
        ValueError, Compounded.FromPrimary, self.connection, (1,))
    self.assertRaises(
        ValueError, Compounded.FromPrimary, self.connection, (1, 2, 3))

  def testKeying(self):
    """[Compound] Compound record raises as expected in case of duplicate key"""
    Compounded.Create(self.connection, {
        'first': 1, 'second': 1, 'message': 'very first'})
    Compounded.Create(self.connection, {
        'first': 1, 'second': 2, 'message': 'second messge'})
    Compounded.Create(self.connection, {
        'first': 2, 'second': 1, 'message': 'three is a charm'})
    self.assertRaises(
        self.connection.IntegrityError, Compounded.Create,
        self.connection, {'first': 2, 'second': 1, 'message': 'Break stuff'})



def DatabaseConnection():
  """Returns an SQLTalk database connection to 'uWeb3_model_test'."""
  return mysql.Connect(
      host='localhost',
      user='stef',
      passwd='24192419',
      db='uweb_test',
      charset='utf8')



if __name__ == '__main__':
  unittest.main(testRunner=unittest.TextTestRunner(verbosity=2))



import sqlite3
import datetime
import pytz
import time


INTERPRET_AS_UTC = pytz.utc.localize

def DateFromTicks(ticks):
  return datetime.date(*time.gmtime(ticks)[:3])


def TimeFromTicks(ticks):
  return INTERPRET_AS_UTC(datetime.time(*time.gmtime(ticks)[3:6]))


def TimestampFromTicks(ticks):
  return INTERPRET_AS_UTC(datetime.datetime.utcfromtimestamp(ticks))


def AdaptDate(date_obj):
  """Adapts a datetime.date object into its daynumber since Common Era."""
  return date_obj.toordinal()


def AdaptDatetime(date_obj):
  """Adapts a datetime.datetime object into a short string timestamp.

  The date portion is converted to its daynumber since Common Era.
  The time portion is converted to milliseconds past midnight.
  These two are joined by a 'T'.
  """
  try:
    date_obj = date_obj.astimezone(pytz.utc)
  except ValueError:
    pass  # naive datetime object
  timepart = date_obj.time()
  seconds = (timepart.hour * 3600000 + timepart.minute * 60000 +
             timepart.second * 1000 + timepart.microsecond // 1000)
  return '%dT%d' % (date_obj.toordinal(), seconds)


def AdaptReadableDate(date_obj):
  """Adapts a datetime.date object to its ISO-8601 date notation."""
  return date_obj.isoformat()


def AdaptReadableDatetime(date_obj):
  """Adapts a datetime.datetime object to its ISO-8601 date/time notation."""
  try:
    date_obj = date_obj.astimezone(pytz.utc)
  except ValueError:
    pass  # naive datetime object
  return date_obj.isoformat()


#TODO(Elmer): Docstrings, tests, sharp eyes.
def AdaptReadableTimeStruct(time_struct):
  return time.strftime('%FT%T', time_struct)


def AdaptTimeStruct(time_struct):
  return AdaptDatetime(datetime.datetime(*time_struct[:6]))


def ConvertDate(date_obj):
  """Converts an SQLite DATE field to a datetime.date object.

  Two stored formats are supported: The format as defined in SQLTalk, which is
  the number of days since Common Era, and the ISO-8601 date format."""
  try:
    # Assume date is the number of days since Common Era as a string.
    return datetime.date.fromordinal(int(date_obj))
  except ValueError:
    # Date was not a number, assume ISO-8601
    return datetime.date(*map(int, date_obj.split("-")))


def ConvertTimestamp(date_obj):
  """Converts an encoded timestamp to a datetime.datetime object.

  This reads both a ISO-8601 formatted string, as well as the Underdark custom
  compressed datetime format as defined in AdaptDatetime above.
  """
  sep = 'T' if 'T' in date_obj else ' '
  try:
    # Assume that the datepart is days since Common Era,
    # and the timepart is seconds past midnight.
    datepart, timepart = date_obj.split(sep)
    return INTERPRET_AS_UTC(
        datetime.datetime.fromordinal(int(datepart)) +
        datetime.timedelta(microseconds=int(timepart) * 1000))
  except ValueError:
    # Either part couldn't be interpreter as integer, meaning they are ISO-8601.
    date_obj, _sep, microseconds = date_obj.partition('.')
    time_tuple = time.strptime(date_obj, '%Y-%m-%d' + sep + '%H:%M:%S')[:6]
    if microseconds:
      microseconds = int((microseconds + '00000')[:6])
      time_tuple += microseconds,
    return INTERPRET_AS_UTC(datetime.datetime(*time_tuple))


sqlite3.register_adapter(datetime.date, AdaptDate)
sqlite3.register_adapter(datetime.datetime, AdaptDatetime)
sqlite3.register_adapter(time.struct_time, AdaptTimeStruct)
sqlite3.register_converter('DATE', ConvertDate)
sqlite3.register_converter('TIMESTAMP', ConvertTimestamp)
